import os

from leapp import reporting
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import IfCfg, InstalledRPM, RpmTransactionTasks

FMT_LIST_SEPARATOR = '\n    - '


def process():
    TRUE_VALUES = ['yes', 'true', '1']
    TYPE_MAP = {
        'ethernet':     'NetworkManager',
        'ctc':          'NetworkManager',
        'infiniband':   'NetworkManager',
        'bond':         'NetworkManager',
        'vlan':         'NetworkManager',
        'bridge':       'NetworkManager',
        'wireless':     'NetworkManager-wifi',
        'team':         'NetworkManager-team',
    }

    bad_type_files = []
    not_controlled_files = []
    rpms_to_install = []

    if not has_package(InstalledRPM, 'network-scripts'):
        # If network-scripts package was not installed,
        # we don't do anything.
        return

    for ifcfg in api.consume(IfCfg):
        bad_type = False
        got_type = None
        nm_controlled = True

        if ifcfg.rules is not None or ifcfg.rules6 is not None:
            if 'NetworkManager-dispatcher-routing-rules' not in rpms_to_install:
                rpms_to_install.append('NetworkManager-dispatcher-routing-rules')
            continue

        if os.path.basename(ifcfg.filename) == 'ifcfg-lo':
            continue

        for prop in ifcfg.properties:
            if prop.name in ('TYPE', 'DEVICETYPE'):
                if got_type is None:
                    got_type = prop.value.lower()
                elif got_type != prop.value.lower():
                    bad_type = True

            if prop.name == 'BONDING_MASTER':
                if got_type is None:
                    got_type = 'bond'
                elif got_type != 'bond':
                    bad_type = True

            if prop.name == 'NM_CONTROLLED' and prop.value.lower() not in TRUE_VALUES:
                nm_controlled = False

        if got_type in TYPE_MAP:
            if TYPE_MAP[got_type] not in rpms_to_install:
                rpms_to_install.append(TYPE_MAP[got_type])
        else:
            bad_type = True

        # Don't bother reporting the file for NM_CONTROLLED=no
        # if its type is not supportable with NetworkManager anyway
        if bad_type is True:
            bad_type_files.append(ifcfg.filename)
        elif nm_controlled is False:
            not_controlled_files.append(ifcfg.filename)

    if bad_type_files:
        title = 'Network configuration for unsupported device types detected'
        summary = ('RHEL 9 does not support the legacy network-scripts'
                   ' package that was deprecated in RHEL 8 in favor of'
                   ' NetworkManager. Files for device types that are not'
                   ' supported by NetworkManager are present in the system.'
                   ' Files with the problematic configuration:{}').format(
            ''.join(['{}{}'.format(FMT_LIST_SEPARATOR, bfile) for bfile in bad_type_files])
        )
        remediation = ('Consult the nm-settings-ifcfg-rh(5) manual for'
                       ' valid types of ifcfg files. Remove configuration'
                       ' files that can not be supported.')
        reporting.create_report([
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Remediation(hint=remediation),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.NETWORK, reporting.Groups.SERVICES]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.RelatedResource('package', 'NetworkManager'),
        ] + [
            reporting.RelatedResource('file', fname)
            for fname in bad_type_files
        ])

    if not_controlled_files:
        title = 'Network configuration with disabled NetworkManager support detected'
        summary = ('RHEL 9 does not support the legacy network-scripts'
                   ' package that was deprecated in RHEL 8 in favor of'
                   ' NetworkManager. Configuration present in the system'
                   ' prohibit NetworkManager from loading it.'
                   ' Files with the problematic configuration:{}').format(
            ''.join(['{}{}'.format(FMT_LIST_SEPARATOR, bfile) for bfile in not_controlled_files])
        )
        remediation = ('Ensure the ifcfg files comply with format described in'
                       ' nm-settings-ifcfg-rh(5) manual and remove the'
                       ' NM_CONTROLLED key from them.')
        reporting.create_report([
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Remediation(hint=remediation),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.NETWORK, reporting.Groups.SERVICES]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.RelatedResource('package', 'NetworkManager'),
            reporting.ExternalLink(
                title='nm-settings-ifcfg-rh - Description of ifcfg-rh settings plugin',
                url='https://red.ht/nm-settings-ifcfg-rh',
            ),
        ] + [
            reporting.RelatedResource('file', fname)
            for fname in not_controlled_files
        ])

    if rpms_to_install:
        if not has_package(InstalledRPM, 'NetworkManager'):
            # If the user was not using NetworkManager previously,
            # make sure NetworkManager is configured consistently with how
            # network-scripts behaved.
            rpms_to_install.append('NetworkManager-config-server')
        api.produce(RpmTransactionTasks(to_install=rpms_to_install))



from __future__ import print_function
import json

from leapp.utils.repository import find_repository_basedir
from leapp.repository.scan import find_and_scan_repositories

base_dir = find_repository_basedir('.')
repository = find_and_scan_repositories(base_dir, include_locals=True)

repository.load()

if not hasattr(repository, 'repos'):
    repository.repos = [repository]

print(json.dumps([repo.serialize() for repo in repository.repos]))



from leapp.models import ActiveKernelModule, ActiveKernelModulesFacts
from leapp.reporting import Report
from leapp.utils.report import is_inhibitor


def test_actor_with_nvidia_driver(current_actor_context):
    with_nvidia = [
        ActiveKernelModule(filename='nvidia', parameters=[]),
        ActiveKernelModule(filename='kvm', parameters=[])]

    current_actor_context.feed(ActiveKernelModulesFacts(kernel_modules=with_nvidia))
    current_actor_context.run()
    report_fields = current_actor_context.consume(Report)[0].report
    assert is_inhibitor(report_fields)


def test_actor_without_nvidia_driver(current_actor_context):
    without_nvidia = [
        ActiveKernelModule(filename='i915', parameters=[]),
        ActiveKernelModule(filename='kvm', parameters=[])]

    current_actor_context.feed(ActiveKernelModulesFacts(kernel_modules=without_nvidia))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_with_nouveau_driver(current_actor_context):
    without_nvidia = [
        ActiveKernelModule(filename='nouveau', parameters=[]),
        ActiveKernelModule(filename='kvm', parameters=[])]

    current_actor_context.feed(ActiveKernelModulesFacts(kernel_modules=without_nvidia))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



import pytest

from leapp.libraries import stdlib
from leapp.libraries.actor import sourcebootloaderscanner
from leapp.libraries.common.testutils import produce_mocked

GRUBBY_INFO_ALL_STDOUT = '''index=0
kernel="/boot/vmlinuz-4.18.0-305.7.1.el8_4.x86_64"
args="ro uned_params"
root="/someroot"
initrd="/boot/initramfs-4.18.0-305.7.1.el8_4.x86_64.img"
title="Linux"
id="some_id"
index=1
kernel="/boot/vmlinuz-4.18.0-305.3.1.el8_4.x86_64"
args="ro"
root="/someroot"
initrd="/boot/initramfs-4.18.0-305.3.1.el8_4.x86_64.img"
title="Linux old-kernel"
id="some_id2"
index=2
non linux entry'''


def test_scan_boot_entries(monkeypatch):
    """Tests whether the library correctly identifies boot entries in the grubby output."""
    def run_mocked(cmd, **kwargs):
        if cmd == ['grubby', '--info', 'ALL']:
            return {
                'stdout': GRUBBY_INFO_ALL_STDOUT.split('\n')
            }
        raise ValueError('Tried to run unexpected command.')

    actor_produces = produce_mocked()

    # The library imports `run` all the way (from ... import run), therefore,
    # we must monkeypatch the reference directly in the actor's library namespace
    monkeypatch.setattr(sourcebootloaderscanner, 'run', run_mocked)
    monkeypatch.setattr(stdlib.api, 'produce', actor_produces)

    sourcebootloaderscanner.scan_source_boot_loader_configuration()

    fail_description = 'Only one SourceBootLoaderConfiguration message should be produced.'
    assert len(actor_produces.model_instances) == 1, fail_description

    bootloader_config = actor_produces.model_instances[0]

    fail_description = 'Found different number of boot entries than present in provided mocks.'
    assert len(bootloader_config.entries) == 2, fail_description

    expected_entries = [
        {'title': 'Linux', 'kernel_image': '/boot/vmlinuz-4.18.0-305.7.1.el8_4.x86_64'},
        {'title': 'Linux old-kernel', 'kernel_image': '/boot/vmlinuz-4.18.0-305.3.1.el8_4.x86_64'},
    ]

    actual_entries = sorted(bootloader_config.entries, key=lambda entry: entry.title)

    for actual_entry, expected_entry in zip(actual_entries, expected_entries):
        assert actual_entry.title == expected_entry['title']
        assert actual_entry.kernel_image == expected_entry['kernel_image']



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class DNFPluginTask(Model):
    """
    Represents information what should DNF do with a specific DNF plugin
    in various stages.

    Currently, it's possible just to disable specified DNF plugins.

    Available stages: "check", "download" and "upgrade
    """

    topic = SystemInfoTopic

    name = fields.String()
    disable_in = fields.List(fields.String(), default=['check', 'download', 'upgrade'])



import errno

from leapp.libraries.common.rpms import check_file_modification
from leapp.libraries.stdlib import api
from leapp.models import OpenSslConfig, OpenSslConfigBlock, OpenSslConfigPair

CONFIG = '/etc/pki/tls/openssl.cnf'


def strip_whitespace_and_comments(line):
    """
    Returns everything before the first hash sign after stripping leading and trailing whitespace
    """
    parts = line.split("#", 1)
    return parts[0].strip()


def parse_config(config):
    """
    Parse openssl.cnf configuration
    """
    ret = OpenSslConfig(blocks=[])

    block = None
    for line in config:
        line = strip_whitespace_and_comments(line)
        if not line:
            continue

        # match the block header
        if line[0] == "[" and line[-1] == "]":
            name = line[1:-1].strip()
            block = OpenSslConfigBlock(name=name, pairs=[])
            ret.blocks.append(block)
            continue

        # the rest values are key-value pairs separated with equal sign
        el = line.split('=', 1)
        if len(el) < 2:
            # The special options .include and .pragma can appear without the equal sign
            if el[0].startswith('.include '):
                key = ".include"
                value = el[0][8:].strip()
            # elif el.startswith('.pragma '):
                # we do not care about this option now
            else:
                continue
        else:
            key = el[0].strip()
            value = el[1].strip()

        if block:
            pair = OpenSslConfigPair(key=key, value=value)
            block.pairs.append(pair)
            continue

        if key == 'openssl_conf':
            ret.openssl_conf = value

    return ret


def produce_config(producer, config):
    """
    Produce a Leapp message with all interesting openssl configuration options.
    """

    producer(config)


def read_config():
    """
    Read the actual configuration file.
    """
    try:
        with open(CONFIG, 'r') as fd:
            return fd.readlines()
    except IOError as err:
        if err.errno != errno.ENOENT:
            error = 'Failed to read config: {}'.format(str(err))
            api.current_logger().error(error)
        return []


def scan_config(producer):
    """
    Parse openssl.cnf file to create OpenSslConfig message.
    """

    # direct access to configuration file
    output = read_config()
    config = parse_config(output)

    # find out whether the file was modified from the one shipped in rpm
    config.modified = check_file_modification(CONFIG)

    produce_config(producer, config)



from leapp.actors import Actor
from leapp.libraries.actor.migratentp import migrate_ntp
from leapp.models import NtpMigrationDecision, Report
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MigrateNtp(Actor):
    """
    Migrate ntp and/or ntpdate configuration to chrony.
    """

    name = 'migrate_ntp'
    consumes = (NtpMigrationDecision,)
    produces = (Report,)
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        for decision in self.consume(NtpMigrationDecision):
            migrate_ntp(decision.migrate_services, decision.config_tgz64)



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class BlackListError(Model):
    """
    Provides an entry for all disabled CAs in one of blacklist directoriesi
    which needs to be moved to blocklist.
    """
    topic = SystemInfoTopic

    sourceDir = fields.String()
    """
    The path of the blacklist directory where distrusted certs reside.
    """

    targetDir = fields.String()
    """
    The path of the blocklist directory where distructed certs should reside.
    """

    error = fields.String()
    """
    Errors string from the OS or the LEAPP run process
    """



from collections import defaultdict
from enum import IntEnum

from leapp import reporting
from leapp.libraries.common.config.version import get_source_major_version, get_target_major_version
from leapp.libraries.stdlib import api
from leapp.models import DetectedDeviceOrDriver


class MessagingClass(IntEnum):
    UNKNOWN = 0
    DRIVERS = 1
    DEVICES = 2
    CPUS = 3


def create_inhibitors(inhibiting_entries):
    if not inhibiting_entries:
        return

    drivers = inhibiting_entries.get(MessagingClass.DRIVERS)
    if drivers:
        reporting.create_report([
            reporting.Title(
                'Leapp detected loaded kernel drivers which have been removed '
                'in RHEL {}. Upgrade cannot proceed.'.format(get_target_major_version())
            ),
            reporting.Summary(
                (
                    'Support for the following RHEL {source} device drivers has been removed in RHEL {target}:\n'
                    '     - {drivers}\n'
                ).format(
                    drivers='\n     - '.join([entry.driver_name for entry in drivers]),
                    target=get_target_major_version(),
                    source=get_source_major_version(),
                )
            ),
            reporting.Audience('sysadmin'),
            reporting.Groups([reporting.Groups.KERNEL, reporting.Groups.DRIVERS]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.INHIBITOR])
        ])

    devices = inhibiting_entries.get(MessagingClass.DEVICES)
    if devices:
        reporting.create_report([
            reporting.Title(
                'Leapp detected devices which are no longer supported in RHEL {}. Upgrade cannot proceed.'.format(
                    get_target_major_version())
            ),
            reporting.Summary(
                (
                    'Support for the following devices has been removed in RHEL {target}:\n'
                    '     - {devices}\n'
                ).format(
                    devices='\n     - '.join(['{name} ({pci})'.format(name=entry.device_name,
                                             pci=entry.device_id) for entry in devices]),
                    target=get_target_major_version(),
                )
            ),
            reporting.Audience('sysadmin'),
            reporting.Groups([reporting.Groups.KERNEL]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.INHIBITOR])
        ])

    cpus = inhibiting_entries.get(MessagingClass.CPUS)
    if cpus:
        reporting.create_report([
            reporting.Title(
                'Leapp detected a processor which is no longer supported in RHEL {}. Upgrade cannot proceed.'.format(
                    get_target_major_version())
            ),
            reporting.Summary(
                (
                    'Support for the following processors has been removed in RHEL {target}:\n'
                    '     - {processors}\n'
                ).format(
                    processors='\n     - '.join([entry.device_name for entry in cpus]),
                    target=get_target_major_version(),
                )
            ),
            reporting.Audience('sysadmin'),
            reporting.Groups([reporting.Groups.KERNEL, reporting.Groups.BOOT]),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.INHIBITOR])
        ])


def create_warnings(unmaintained_entries):
    if not unmaintained_entries:
        return

    drivers = unmaintained_entries.get(MessagingClass.DRIVERS)
    if drivers:
        reporting.create_report([
            reporting.Title(
                'Leapp detected loaded kernel drivers which are no longer maintained in RHEL {}.'.format(
                    get_target_major_version())
            ),
            reporting.Summary(
                (
                    'The following RHEL {source} device drivers are no longer maintained RHEL {target}:\n'
                    '     - {drivers}\n'
                ).format(
                    drivers='\n     - '.join([entry.driver_name for entry in drivers]),
                    target=get_target_major_version(),
                    source=get_source_major_version(),
                )
            ),
            reporting.Audience('sysadmin'),
            reporting.Groups([reporting.Groups.KERNEL, reporting.Groups.DRIVERS]),
            reporting.Severity(reporting.Severity.HIGH),
        ])

    devices = unmaintained_entries.get(MessagingClass.DEVICES)
    if devices:
        reporting.create_report([
            reporting.Title(
                'Leapp detected devices which are no longer maintained in RHEL {}'.format(
                    get_target_major_version())
            ),
            reporting.Summary(
                (
                    'The support for the following devices has been removed in RHEL {target} and '
                    'are no longer maintained:\n     - {devices}\n'
                ).format(
                    devices='\n     - '.join(['{name} ({pci})'.format(name=entry.device_name,
                                             pci=entry.device_id) for entry in devices]),
                    target=get_target_major_version(),
                )
            ),
            reporting.Audience('sysadmin'),
            reporting.Groups([reporting.Groups.KERNEL]),
            reporting.Severity(reporting.Severity.HIGH),
        ])

    cpus = unmaintained_entries.get(MessagingClass.CPUS)
    if cpus:
        reporting.create_report([
            reporting.Title(
                'Leapp detected a processor which is no longer maintained in RHEL {}.'.format(
                    get_target_major_version())
            ),
            reporting.Summary(
                (
                    'The following processors are no longer maintained in RHEL {target}:\n'
                    '     - {processors}\n'
                ).format(
                    processors='\n     - '.join([entry.device_name for entry in cpus]),
                    target=get_target_major_version(),
                )
            ),
            reporting.Audience('sysadmin'),
            reporting.Groups([reporting.Groups.KERNEL, reporting.Groups.BOOT]),
            reporting.Severity(reporting.Severity.HIGH),
        ])


def classify(entry):
    if entry.device_type == 'pci':
        if entry.device_id:
            return MessagingClass.DEVICES
        return MessagingClass.DRIVERS
    if entry.device_type == 'cpu':
        return MessagingClass.CPUS
    return MessagingClass.UNKNOWN


def process():
    target_version = int(get_target_major_version())
    inhibiting = defaultdict(list)
    unmaintained = defaultdict(list)
    for entry in api.consume(DetectedDeviceOrDriver):
        if target_version not in entry.available_in_rhel:
            inhibiting[classify(entry)].append(entry)
        elif target_version not in entry.maintained_in_rhel:
            unmaintained[classify(entry)].append(entry)
    create_inhibitors(inhibiting)
    create_warnings(unmaintained)



import os


def is_ipa_client_configured():
    """
    Check if ipa-client is configured on the system

    :return: True if client is configured
    :rtype: bool
    """
    return all(
        (
            os.path.isfile("/etc/ipa/default.conf"),
            os.path.isfile("/var/lib/ipa-client/sysrestore/sysrestore.state"),
        )
    )


def is_ipa_server_configured():
    """
    Check if ipa-server is configured on the system

    :return: True if server is configured
    :rtype: bool
    """
    return all(
        (
            is_ipa_client_configured(),
            os.path.isfile("/var/lib/ipa/sysrestore/sysrestore.state"),
        )
    )



import json
import os
import re
import shutil
import tempfile

from six.moves import urllib

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.common.gpg import get_gpg_fp_from_file, get_path_to_gpg_certs, is_nogpgcheck_set
from leapp.libraries.stdlib import api
from leapp.models import (
    DNFWorkaround,
    TargetUserSpaceInfo,
    TMPTargetRepositoriesFacts,
    TrustedGpgKeys,
    UsedTargetRepositories
)
from leapp.utils.deprecation import suppress_deprecation

FMT_LIST_SEPARATOR = '\n    - '


def _expand_vars(path):
    """
    Expand variables like $releasever and $basearch to the target system version
    """
    r = path.replace('$releasever', get_target_major_version())
    r = r.replace('$basearch', api.current_actor().configuration.architecture)
    return r


def _get_abs_file_path(target_userspace, file_url):
    """
    Return the absolute path for file_url if starts with file:///

    If the file_url starts with 'file:///', return its absolute path to
    the target userspace container, as such a file is supposed to be located
    on the target system. If the path does not exist in the container, the
    the path to the source OS filesystem is returned regardless it exists or not.

    For all other cases, return the originally obtained value.
    """
    if not isinstance(target_userspace, TargetUserSpaceInfo):
        # not need to cover this by tests, it's seatbelt
        raise ValueError('target_userspace must by TargetUserSpaceInfo object')

    prefix = 'file:///'
    if not file_url.startswith(prefix):
        return file_url

    file_path = file_url[len(prefix):]
    expanded = os.path.join(target_userspace.path, file_path)
    if os.path.exists(expanded):
        return expanded

    # the file does not exist in the container -- try the path in the source OS
    return os.path.join('/', file_path)


def _consume_data():
    try:
        used_target_repos = next(api.consume(UsedTargetRepositories)).repos
    except StopIteration:
        raise StopActorExecutionError(
            'Could not check for valid GPG keys', details={'details': 'No UsedTargetRepositories facts'}
        )

    try:
        target_repos = next(api.consume(TMPTargetRepositoriesFacts)).repositories
    except StopIteration:
        raise StopActorExecutionError(
            'Could not check for valid GPG keys', details={'details': 'No TMPTargetRepositoriesFacts facts'}
        )
    try:
        trusted_gpg_keys = next(api.consume(TrustedGpgKeys))
    except StopIteration:
        raise StopActorExecutionError(
            'Could not check for valid GPG keys', details={'details': 'No TrustedGpgKeys facts'}
        )
    try:
        target_userspace = next(api.consume(TargetUserSpaceInfo))
    except StopIteration:
        raise StopActorExecutionError(
            'Could not check for valid GPG keys', details={'details': 'No TargetUserSpaceInfo facts'}
        )

    return used_target_repos, target_repos, trusted_gpg_keys, target_userspace


def _get_repo_gpgkey_urls(repo):
    """
    Return the list or repository gpgkeys that should be checked

    If the gpgcheck is disabled for the repo or gpgkey is not specified,
    return an empty list.

    Returned gpgkeys are URLs with already expanded variables
    (e.g. $releasever) as gpgkey can contain list of URLs separated by comma
    or whitespaces.
    If gpgcheck=0 is present in the repo file, [] is returned. If the
    gpgcheck is missing or enabled and no gpgkey is present, None is
    returned, which means the repo can not be checked.
    """

    if not repo.additional_fields:
        return None

    repo_additional = json.loads(repo.additional_fields)

    # TODO does the case matter here?
    if 'gpgcheck' in repo_additional and repo_additional['gpgcheck'] in ('0', 'False', 'no'):
        # NOTE: https://dnf.readthedocs.io/en/latest/conf_ref.html#boolean-label
        # nothing to do with repos with enforced gpgcheck=0
        return []

    if 'gpgkey' not in repo_additional:
        # This means rpm will bail out at some time if the key is not present
        # but we will not know if the needed key is present or not before we will have
        # the packages at least downloaded
        api.current_logger().warning(
            'The gpgcheck for the {} repository is enabled'
            ' but gpgkey is not specified. Cannot be checked.'
            .format(repo.repoid)
        )
        return None

    return re.findall(r'[^,\s]+', _expand_vars(repo_additional['gpgkey']))


def _report(title, summary, keys, inhibitor=False):
    summary = (
        '{summary}'
        ' Leapp is not able to guarantee validity of such gpg keys and manual'
        ' review is required, so any spurious keys are not imported in the system'
        ' during the in-place upgrade.'
        ' The following additional gpg keys are required to be imported during'
        ' the upgrade:{sep}{key_list}'
        .format(
            summary=summary,
            sep=FMT_LIST_SEPARATOR,
            key_list=FMT_LIST_SEPARATOR.join(keys)
        )
    )
    hint = (
        'Check the path to the listed GPG keys is correct, the keys are valid and'
        ' import them into the host RPM DB or store them inside the {} directory'
        ' prior the upgrade.'
        ' If you want to proceed the in-place upgrade without checking any RPM'
        ' signatures, execute leapp with the `--nogpgcheck` option.'
        .format(get_path_to_gpg_certs())
    )
    groups = [reporting.Groups.REPOSITORY]
    if inhibitor:
        groups.append(reporting.Groups.INHIBITOR)
    reporting.create_report(
        [
            reporting.Title(title),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups(groups),
            reporting.Remediation(hint=hint),
            # TODO(pstodulk): @Jakuje: let's sync about it
            # TODO update external documentation ?
            # reporting.ExternalLink(
            #     title=(
            #         "Customizing your Red Hat Enterprise Linux "
            #         "in-place upgrade"
            #     ),
            #     url=(
            #         "https://access.redhat.com/articles/4977891/"
            #         "#repos-known-issues"
            #     ),
            # ),
        ]
    )


def _report_missing_keys(keys):
    summary = (
        'Some of the target repositories require GPG keys that are not installed'
        ' in the current RPM DB or are not stored in the {trust_dir} directory.'
        .format(trust_dir=get_path_to_gpg_certs())
    )
    _report('Detected unknown GPG keys for target system repositories', summary, keys, True)


def _report_failed_download(keys):
    summary = (
        'Some of the target repositories require GPG keys that are referenced'
        ' using remote protocol (http:// or https://) but can not be downloaded.'
    )
    _report('Failed to download GPG key for target repository', summary, keys)


def _report_unknown_protocol(keys):
    summary = (
        'Some of the target repositories require GPG keys that are provided'
        ' using unknown protocol.'
    )
    _report('GPG keys provided using unknown protocol', summary, keys)


def _report_invalid_keys(keys):
    summary = (
        'Some of the target repositories require GPG keys, which point to files'
        ' that do not contain any gpg keys.'
    )
    _report('Failed to read GPG keys from provided key files', summary, keys)


def _report_repos_missing_keys(repos):
    summary = (
        'Some of the target repositories require checking GPG signatures, but do'
        ' not provide any gpg keys.'
        ' Leapp is not able to guarantee validity of such gpg keys and manual'
        ' review is required, so any spurious keys are not imported in the system'
        ' during the in-place upgrade.'
        ' The following repositories require some attention before the upgrade:'
        ' {sep}{key_list}'
        .format(
            sep=FMT_LIST_SEPARATOR,
            key_list=FMT_LIST_SEPARATOR.join(repos)
        )
    )
    hint = (
        'Check the repositories are correct and either add a respective gpgkey='
        ' option, disable checking RPM signature using gpgcheck=0 per-repository.'
        ' If you want to proceed the in-place upgrade without checking any RPM'
        ' signatures, execute leapp with the `--nogpgcheck` option.'
    )
    groups = [reporting.Groups.REPOSITORY]
    reporting.create_report(
        [
            reporting.Title('Inconsistent repository without GPG key'),
            reporting.Summary(summary),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups(groups),
            reporting.Remediation(hint=hint),
            # TODO(pstodulk): @Jakuje: let's sync about it
            # TODO update external documentation ?
            # reporting.ExternalLink(
            #     title=(
            #         "Customizing your Red Hat Enterprise Linux "
            #         "in-place upgrade"
            #     ),
            #     url=(
            #         "https://access.redhat.com/articles/4977891/"
            #         "#repos-known-issues"
            #     ),
            # ),
        ]
    )


def register_dnfworkaround():
    api.produce(DNFWorkaround(
        display_name='import trusted gpg keys to RPM DB',
        script_path=api.current_actor().get_common_tool_path('importrpmgpgkeys'),
        script_args=[get_path_to_gpg_certs()],
    ))


@suppress_deprecation(TMPTargetRepositoriesFacts)
def process():
    """
    Process the repositories and find missing signing keys

    UsedTargetRepositories doesn't contain baseurl attribute. So gathering
    them from model TMPTargetRepositoriesFacts.
    """
    # when the user decided to ignore gpg signatures on the packages, we can ignore these checks altogether
    if is_nogpgcheck_set():
        api.current_logger().warning('The --nogpgcheck option is used: skipping all related checks.')
        return

    used_target_repos, target_repos, trusted_gpg_keys, target_userspace = _consume_data()

    target_repo_id_to_repositories_facts_map = {
        repo.repoid: repo
        for repofile in target_repos
        for repo in repofile.data
    }

    # For reporting all the issues in one batch instead of reporting each issue in separate report
    missing_keys = list()
    failed_download = list()
    unknown_protocol = list()
    invalid_keys = list()
    repos_missing_keys = list()

    pubkeys = [key.fingerprint for key in trusted_gpg_keys.items]
    processed_gpgkey_urls = set()
    tmpdir = None
    for repoid in used_target_repos:
        if repoid.repoid not in target_repo_id_to_repositories_facts_map:
            api.current_logger().warning('The target repository {} metadata not available'.format(repoid.repoid))
            continue

        repo = target_repo_id_to_repositories_facts_map[repoid.repoid]
        gpgkeys = _get_repo_gpgkey_urls(repo)
        if gpgkeys is None:
            repos_missing_keys.append(repo.repoid)
            continue
        for gpgkey_url in gpgkeys:
            if gpgkey_url in processed_gpgkey_urls:
                continue
            processed_gpgkey_urls.add(gpgkey_url)

            if gpgkey_url.startswith('file:///'):
                key_file = _get_abs_file_path(target_userspace, gpgkey_url)
            elif gpgkey_url.startswith('http://') or gpgkey_url.startswith('https://'):
                # delay creating temporary directory until we need it
                tmpdir = tempfile.mkdtemp() if tmpdir is None else tmpdir
                # FIXME: what to do with dummy? it's fd, that should be closed also
                dummy, tmp_file = tempfile.mkstemp(dir=tmpdir)
                try:
                    urllib.request.urlretrieve(gpgkey_url, tmp_file)
                    key_file = tmp_file
                except urllib.error.URLError as err:
                    api.current_logger().warning(
                        'Failed to download the gpgkey {}: {}'.format(gpgkey_url, str(err)))
                    failed_download.append(gpgkey_url)
                    continue
            else:
                unknown_protocol.append(gpgkey_url)
                api.current_logger().error(
                    'Skipping unknown protocol for gpgkey {}'.format(gpgkey_url))
                continue
            fps = get_gpg_fp_from_file(key_file)
            if not fps:
                invalid_keys.append(gpgkey_url)
                api.current_logger().warning(
                    'Cannot get any gpg key from the file: {}'.format(gpgkey_url)
                )
                continue
            for fp in fps:
                if fp not in pubkeys and gpgkey_url not in missing_keys:
                    missing_keys.append(_get_abs_file_path(target_userspace, gpgkey_url))

    if tmpdir:
        # clean up temporary directory with downloaded gpg keys
        shutil.rmtree(tmpdir)

    # report
    if failed_download:
        _report_failed_download(failed_download)
    if unknown_protocol:
        _report_unknown_protocol(unknown_protocol)
    if invalid_keys:
        _report_invalid_keys(invalid_keys)
    if missing_keys:
        _report_missing_keys(missing_keys)
    if repos_missing_keys:
        _report_repos_missing_keys(repos_missing_keys)

    register_dnfworkaround()



from leapp.actors import Actor
from leapp.libraries.actor import checkblacklistca
from leapp.models import BlackListCA, BlackListError, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckBlackListCA(Actor):
    """
    No documentation has been provided for the checkblacklistca actor.
    """

    name = 'checkblacklistca'
    consumes = (BlackListCA, BlackListError)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        checkblacklistca.process()



import os

from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import kernelcmdlineconfig
from leapp.models import FirmwareFacts, InstalledTargetKernelInfo, KernelCmdlineArg, TargetKernelCmdlineArgTasks
from leapp.tags import FinalizationPhaseTag, IPUWorkflowTag


class KernelCmdlineConfig(Actor):
    """
    Append extra arguments to the target RHEL kernel command line
    """

    name = 'kernelcmdlineconfig'
    consumes = (KernelCmdlineArg, InstalledTargetKernelInfo, FirmwareFacts, TargetKernelCmdlineArgTasks)
    produces = ()
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):

        configs = None
        ff = next(self.consume(FirmwareFacts), None)
        if not ff:
            raise StopActorExecutionError(
                'Could not identify system firmware',
                details={'details': 'Actor did not receive FirmwareFacts message.'}
            )

        if ff.firmware == 'bios' and os.path.ismount('/boot/efi'):
            configs = ['/boot/grub2/grub.cfg', '/boot/efi/EFI/redhat/grub.cfg']
        kernelcmdlineconfig.modify_kernel_args_in_boot_cfg(configs)



from leapp.models import Report, SkippedRepositories


def test_skipped_repos(current_actor_context):
    reported_packages = ['pkg_a', 'pkg_b', 'pkg_c']
    reported_repos = ['repo_a', 'repo_b', 'repo_c']
    current_actor_context.feed(
        SkippedRepositories(
            packages=list(reported_packages),
            repos=list(reported_repos)
        )
    )

    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports
    assert len(reports) == 1
    report_fields = reports[0].report
    for pkg in reported_packages:
        assert '\n- {}'.format(pkg) in report_fields['summary']
    for repo in reported_repos:
        assert '\n- {}'.format(repo) in report_fields['summary']


def test_skipped_just_repos(current_actor_context):
    reported_repos = ['repo_a', 'repo_b', 'repo_c']
    current_actor_context.feed(
        SkippedRepositories(
            packages=[],
            repos=list(reported_repos)
        )
    )

    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert reports
    assert len(reports) == 1
    report_fields = reports[0].report
    for repo in reported_repos:
        assert '\n- {}'.format(repo) in report_fields['summary']


def test_skipped_repos_empty(current_actor_context):
    current_actor_context.feed(
        SkippedRepositories(
            packages=[],
            repos=[]
        )
    )
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert not reports


def test_skipped_repos_no_repos(current_actor_context):
    current_actor_context.feed(
        SkippedRepositories(
            packages=['woot'],
            repos=[]
        )
    )
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert not reports


def test_skipped_repos_no_message(current_actor_context):
    current_actor_context.run()
    reports = list(current_actor_context.consume(Report))
    assert not reports



from leapp.actors import Actor
from leapp.libraries.actor import scancustommodifications
from leapp.models import CustomModifications
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanCustomModificationsActor(Actor):
    """
    Collects information about files in leapp directories that have been modified or newly added.
    """

    name = 'scan_custom_modifications_actor'
    produces = (CustomModifications,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        for msg in scancustommodifications.scan():
            self.produce(msg)



from leapp.actors import Actor
from leapp.libraries.actor import deviceanddriverdeprecationdataload
from leapp.models import ConsumedDataAsset, DeviceDriverDeprecationData
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class LoadDeviceDriverDeprecationData(Actor):
    """
    Loads deprecation data for drivers and devices (PCI & CPU)

    The data will either be loaded from the local /etc/leapp/files location or
    fetched from the Red Hat remote service providing this data.
    """

    name = 'load_device_driver_deprecation_data'
    consumes = ()
    produces = (DeviceDriverDeprecationData, ConsumedDataAsset)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self, *args, **kwargs):
        deviceanddriverdeprecationdataload.process()



import os

from leapp.libraries.stdlib import api, run
from leapp.models import SCTPConfig


def enable_sctp(_modprobe_d_path="/etc/modprobe.d"):
    """
    Enables the SCTP module by removing it from being black listed.
    :type _modprobe_d_path: str
    :param _modprobe_d_path: overwrite only in case of testing, by passing
        some tmp_dir instead
    """

    api.current_logger().info('Enabling SCTP.')
    run(['/usr/bin/sed', '-i', r's/^\s*blacklist.*sctp/#&/',
         os.path.join(_modprobe_d_path, 'sctp_diag-blacklist.conf'),
         os.path.join(_modprobe_d_path, 'sctp-blacklist.conf')])
    api.current_logger().info('Enabled SCTP.')


def perform_update():
    for sctpconfig in api.consume(SCTPConfig):
        api.current_logger().info('Consuming sctp={}'.format(sctpconfig.wanted))
        if sctpconfig.wanted:
            enable_sctp()
            break



import errno
import os

import pytest


@pytest.mark.skipif(
    not os.path.isdir('/etc/systemd/system/default.target.wants')
    or not os.getuid() == 0,
    reason='default.target.wants dir should exists and test should be run '
           'under the root user.',
)
# TODO make the test not destructive
@pytest.mark.skipif(os.getenv("DESTRUCTIVE_TESTING", False) in [False, "0"],
                    reason='Test disabled by default because it would modify the system')
def test_remove_resume_service(current_actor_context):
    service_name = 'leapp_resume.service'
    service_path = os.path.join('/etc/systemd/system/', service_name)
    symlink_path = os.path.join('/etc/systemd/system/default.target.wants/', service_name)

    # lets make sure there are not leftovers from previous tests
    try:
        os.unlink(service_path)
        os.unlink(symlink_path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise

    with open(service_path, 'w'):
        pass

    os.symlink(service_path, symlink_path)

    current_actor_context.run()

    assert not os.path.isfile(service_path)
    assert not os.path.isfile(symlink_path)



from leapp import reporting
from leapp.libraries.common.config import architecture


def check_architecture():
    """Check if given architecture is supported by upgrade process"""
    if not architecture.matches_architecture(*architecture.ARCH_SUPPORTED):
        inhibit_upgrade()


def inhibit_upgrade():
    """Generate an upgrade inhibitor"""
    reporting.create_report(
        [
            reporting.Title('Unsupported architecture'),
            reporting.Summary(
                'Upgrade process is only supported on {} systems.'.format(
                    ', '.join(architecture.ARCH_SUPPORTED)
                )
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SANITY]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
        ]
    )



import pytest

from leapp.libraries.actor.cupsmigrate import migrate_configuration
from leapp.models import CupsChangedFeatures


class MockFileSystem(object):
    def __init__(self,
                 infiles):
        self.files = {}
        self.ssl_dir = []
        self.files = infiles
        for path in infiles.keys():
            if path.startswith('ssl') or path.startswith('/etc/cups/ssl'):
                self.ssl_dir.append(path.rsplit('/', 1)[1])

    def readlines(self, path):
        if path in self.files.keys():
            return self.files[path].splitlines(True)
        raise IOError('Error when reading file {} - file '
                      'does not exist.'.format(path))

    def write(self, path, mode, content):
        if isinstance(content, list):
            content = ''.join(content)

        if mode == 'w':
            self.files[path] = content
        else:
            self.files[path] += content

    def copy_to_ssl(self, oldpath):
        self.ssl_dir.append(oldpath.rsplit('/', 1)[1])


class MockLogger(object):
    def __init__(self):
        self.debug_msg = ''
        self.error_msg = ''

    def debug_log(self, msg):
        self.debug_msg += msg

    def error_log(self, msg):
        self.error_msg += msg


class MockModel(object):
    def __init__(self, facts):
        if not facts:
            self.model = None
            return

        self.model = CupsChangedFeatures(include=facts.get('include', False),
                                         digest=facts.get('digest', False),
                                         env=facts.get('env', False),
                                         certkey=facts.get('certkey', False),
                                         printcap=facts.get('printcap', False),
                                         include_files=facts.get('include_files', []))

    def get_facts(self, model):
        return self.model


testdata = (
    (
        None,
        {
            'debug_msg': '',
            'error_msg': ''
        }
    ),
    (
        {},
        {
            'debug_msg': '',
            'error_msg': ''
        }
    ),
    (
        {
            'include': True,
            'include_files': ['/etc/cups/cupsd.conf', 'smth.conf',
                              'any.conf'],
        },
        {
            'debug_msg': 'Migrating CUPS configuration - Include directives.',
            'error_msg': 'Following errors happened during CUPS migration:\n   '
                         '- Include directive: Error when reading file /etc/cup'
                         's/cupsd.conf - file does not exist.\n   - Include dir'
                         'ective: Error when reading file smth.conf - file does'
                         ' not exist.\n   - Include directive: Error when readi'
                         'ng file any.conf - file does not exist.'
        }
    ),
    (
        {
            'digest': True,
        },
        {
            'debug_msg': 'Migrating CUPS configuration - BasicDigest/Digest'
                         ' directives.',
            'error_msg': 'Following errors happened during CUPS migration:\n   '
                         '- Digest/BasicDigest values: Error when reading file '
                         '/etc/cups/cupsd.conf - file does not exist.'
        }
    ),
    (
        {
            'env': True,
        },
        {
            'debug_msg': 'Migrating CUPS configuration - PassEnv/SetEnv directives.',
            'error_msg': 'Following errors happened during CUPS migration:\n   '
                         '- PassEnv/SetEnv directives: Error when reading file '
                         '/etc/cups/cupsd.conf - file does not exist.'
        }
    ),
    (
        {
            'certkey': True,
        },
        {
            'debug_msg': 'Migrating CUPS configuration - ServerKey/ServerCertif'
                         'icate directive.',
            'error_msg': 'Following errors happened during CUPS migration:\n   '
                         '- ServerKey/ServerCertificate directives: Error when '
                         'reading file /etc/cups/cups-files.conf - file does no'
                         't exist.'
        }
    ),
    (
        {
            'printcap': True,
        },
        {
            'debug_msg': 'Migrating CUPS configuration - PrintcapFormat directive.',
            'error_msg': 'Following errors happened during CUPS migration:\n   '
                         '- PrintcapFormat directive: Error when reading file /'
                         'etc/cups/cupsd.conf - file does not exist.'
        }
    ),
    (
        {
            'certkey': True,
            'include': True,
            'env': True,
            'printcap': True,
            'digest': True,
            'include_files': ['/etc/cups/cupsd.conf', 'smth.conf',
                              'any.conf']
        },
        {
            'debug_msg': 'Migrating CUPS configuration - Include directives.Mig'
                         'rating CUPS configuration - BasicDigest/Digest direct'
                         'ives.Migrating CUPS configuration - PassEnv/SetEnv di'
                         'rectives.Migrating CUPS configuration - ServerKey/Ser'
                         'verCertificate directive.Migrating CUPS configuration'
                         ' - PrintcapFormat directive.',
            'error_msg': 'Following errors happened during CUPS migration:\n   '
                         '- Include directive: Error when reading file /etc/cup'
                         's/cupsd.conf - file does not exist.\n   - Include dir'
                         'ective: Error when reading file smth.conf - file does'
                         ' not exist.\n   - Include directive: Error when readi'
                         'ng file any.conf - file does not exist.\n   - Digest/'
                         'BasicDigest values: Error when reading file /etc/cups'
                         '/cupsd.conf - file does not exist.\n   - PassEnv/SetE'
                         'nv directives: Error when reading file /etc/cups/cups'
                         'd.conf - file does not exist.\n   - ServerKey/ServerC'
                         'ertificate directives: Error when reading file /etc/c'
                         'ups/cups-files.conf - file does not exist.\n   - Prin'
                         'tcapFormat directive: Error when reading file /etc/cu'
                         'ps/cupsd.conf - file does not exist.'
        }
    )
)


@pytest.mark.parametrize('facts,expected', testdata)
def test_migrate_configuration(facts, expected):
    data_model = MockModel(facts)

    op = MockFileSystem({})

    logger = MockLogger()

    migrate_configuration(logger.error_log,
                          logger.debug_log,
                          op,
                          data_model.get_facts)

    assert logger.debug_msg == expected['debug_msg']
    assert logger.error_msg == expected['error_msg']



from leapp.actors import Actor
from leapp.libraries.actor import repositoriesblacklist
from leapp.models import CustomTargetRepository, RepositoriesBlacklisted, RepositoriesFacts, RepositoriesMapping
from leapp.reporting import Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class RepositoriesBlacklist(Actor):
    """
    Exclude target repositories provided by Red Hat without support.

    Conditions to exclude:
    - there are not such repositories already enabled on the source system
      (e.g. "Optional" repositories)
    - such repositories are not required for the upgrade explicitly by the user
      (e.g. via the --enablerepo option or via the /etc/leapp/files/leapp_upgrade_repositories.repo file)

    E.g. CRB repository is provided by Red Hat but it is without the support.
    """

    name = "repositories_blacklist"
    consumes = (
        CustomTargetRepository,
        RepositoriesFacts,
        RepositoriesMapping,
    )
    produces = (RepositoriesBlacklisted, Report)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        repositoriesblacklist.process()



from leapp.actors import Actor
from leapp.libraries.actor import cupsmigrate
from leapp.models import CupsChangedFeatures
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class CupsMigrate(Actor):
    """
    cups_migrate actor

    Migrates configuration directives and writes into error log
    if any error was encountered.
    """

    name = 'cups_migrate'
    consumes = (CupsChangedFeatures,)
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        cupsmigrate.migrate_configuration()



from leapp.actors import Actor
from leapp.models import DNFWorkaround
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class RegisterRubyIRBAdjustment(Actor):
    """
    Registers a workaround which will adjust the Ruby IRB directories during the upgrade.
    """

    name = 'register_ruby_irb_adjustment'
    consumes = ()
    produces = (DNFWorkaround,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        self.produce(
            DNFWorkaround(
                display_name='IRB directory fix',
                script_path=self.get_tool_path('handlerubyirbsymlink'),
            )
        )



from leapp.libraries.common import persistentnetnames
from leapp.libraries.common.testutils import produce_mocked
from leapp.libraries.stdlib import api


class AttributesTest(object):
    def __init__(self):
        self.attributes = {
            'address': b'fa:16:3e:cd:26:5a'
        }

    def get(self, attribute):
        if attribute in self.attributes:
            return self.attributes[attribute]
        raise KeyError


class DeviceTest(object):
    def __init__(self):
        self.dict_data = {
             'ID_NET_DRIVER': 'virtio_net',
             'ID_VENDOR_ID': '0x1af4',
             'ID_PATH': 'pci-0000:00:03.0',
        }

    def __getitem__(self, key):
        if key in self.dict_data:
            return self.dict_data[key]
        raise KeyError

    @property
    def sys_name(self):
        return 'eth'

    @property
    def device_path(self):
        return '/devices/pci0000:00/0000:00:03.0/virtio0/net/eth0'

    @property
    def attributes(self):
        return AttributesTest()


def provide_test_interfaces():
    return [DeviceTest()]


def test_getting_interfaces(monkeypatch):
    monkeypatch.setattr(persistentnetnames, 'physical_interfaces', provide_test_interfaces)
    monkeypatch.setattr(api, 'produce', produce_mocked())
    interface = next(persistentnetnames.interfaces())
    assert interface.name
    assert interface.devpath
    assert interface.driver
    assert interface.vendor
    assert interface.pci_info
    assert interface.mac



import os.path
import re
from collections import namedtuple

import pytest

from leapp.libraries.actor import modscan
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.utils.deprecation import suppress_deprecation

from leapp.models import (  # isort:skip
    RequiredUpgradeInitramPackages,  # deprecated
    UpgradeDracutModule,  # deprecated
    TargetUserSpaceUpgradeTasks,
    UpgradeInitramfsTasks
)


def _files_get_folder_path(name):
    path = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'files', name)
    assert os.path.isdir(path)
    return path


@suppress_deprecation(UpgradeDracutModule)
def test_created_modules(monkeypatch):
    monkeypatch.setattr(api, 'get_actor_folder_path', _files_get_folder_path)
    path = os.path.abspath(api.get_actor_folder_path('dracut'))
    required_modules = {'sys-upgrade', 'sys-upgrade-redhat'}
    # _old is for deprecated stuff, _new is ...
    # we want to test that the new and original solution produce 'equivalent' output
    produced_required_modules_old = set()
    produced_required_modules_new = set()

    def check_module(mod):
        # Ensures that this actor only includes known modules
        assert mod.name in required_modules

        # Ensures that the path is valid
        assert mod.module_path
        assert mod.module_path == os.path.abspath(mod.module_path)

        # Ensure it's a directory
        assert os.path.isdir(mod.module_path)

        # Ensure it's located within the actors files path
        assert mod.module_path.startswith(path)

        # Ensure the directory name ends with the module name
        assert os.path.basename(mod.module_path).endswith(mod.name)
        assert not re.sub(r'^(85|90){}$'.format(mod.name), '', os.path.basename(mod.module_path))

    for msg in modscan._create_dracut_modules():
        if isinstance(msg, UpgradeDracutModule):
            produced_required_modules_old.add(msg.name)
            check_module(msg)
        else:
            for mod in msg.include_dracut_modules:
                produced_required_modules_new.add(mod.name)
                check_module(mod)

    # ensure that old and new solution produce equivalent (expected) output
    assert produced_required_modules_new == produced_required_modules_old == required_modules


@pytest.mark.parametrize('src_ver,dst_ver,arch', (
    ('7.9', '8.4', architecture.ARCH_X86_64),
    ('7.9', '8.4', architecture.ARCH_S390X),
    ('8.4', '9.0', architecture.ARCH_X86_64),
    ('8.4', '9.0', architecture.ARCH_S390X),
))
def test_required_packages(monkeypatch, src_ver, dst_ver, arch):
    # the default set of required rpms should not contain biosdevname
    for pkg in ['biosdevname', 'policycoreutils', 'rng-tools']:
        assert pkg not in modscan._REQUIRED_PACKAGES

    required_packages = modscan._REQUIRED_PACKAGES[:]
    if dst_ver[0] == '9':
        required_packages += ['policycoreutils', 'rng-tools']
    if arch == architecture.ARCH_X86_64:
        required_packages += ['biosdevname']

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver=src_ver, dst_ver=dst_ver, arch=arch))
    old_initram, new_initram = modscan._create_initram_packages()
    assert (set(required_packages)
            == set(old_initram.packages)
            == set(new_initram.install_rpms))


@suppress_deprecation(UpgradeDracutModule)
def test_process_produces_modules(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    messages = []
    monkeypatch.setattr(api, 'produce', lambda *x: messages.extend(x))
    monkeypatch.setattr(api, 'get_actor_folder_path', _files_get_folder_path)
    modscan.process()
    assert messages
    assert len(messages) == 6
    assert len([msg for msg in messages if isinstance(msg, UpgradeDracutModule)]) == 2
    assert len([msg for msg in messages if isinstance(msg, UpgradeInitramfsTasks)]) == 2
    assert len([msg for msg in messages if isinstance(msg, RequiredUpgradeInitramPackages)]) == 1
    assert len([msg for msg in messages if isinstance(msg, TargetUserSpaceUpgradeTasks)]) == 1



import os

from leapp import reporting
from leapp.libraries.actor import trustedgpgkeys
from leapp.libraries.common.gpg import get_pubkeys_from_rpms
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, logger_mocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import GpgKey, InstalledRPM, RPM, TrustedGpgKeys


def _get_test_installed_rmps(fps):
    # adding at least one rpm that is not gpg-pubkey
    rpms = [RPM(
        name='rpm',
        version='4.17.1',
        release='3.fc35',
        epoch='0',
        packager='Fedora Project',
        arch='x86_64',
        pgpsig='RSA/SHA256, Tue 02 Aug 2022 03:12:43 PM CEST, Key ID db4639719867c58f'
    )]
    for fp in fps:
        rpms.append(RPM(
            name='gpg-pubkey',
            version=fp,
            release='5e3006fb',
            epoch='0',
            packager='Fedora (33) <fedora-33-primary@fedoraproject.org>',
            arch='noarch',
            pgpsig=''
        ))
    return InstalledRPM(items=rpms)


class MockedGetGpgFromFile(object):
    def __init__(self, file_fps_tuples):
        # e.g. file_fps_tuple = [('/mydir/myfile', ['0000ff31', '0000ff32'])]
        self._data = {}
        for fname, fps in file_fps_tuples:
            self._data[fname] = fps

    def get_files(self):
        return self._data.keys()  # noqa: W1655; pylint: disable=dict-keys-not-iterating

    def __call__(self, fname):
        return self._data.get(fname, [])


def test_get_pubkeys(monkeypatch):
    """
    Very basic test of _get_pubkeys function
    """
    rpm_fps = ['9570ff31', '99900000']
    file_fps = ['0000ff31', '0000ff32']
    installed_rpms = _get_test_installed_rmps(rpm_fps)
    mocked_gpg_files = MockedGetGpgFromFile([('/mydir/myfile', ['0000ff31', '0000ff32'])])

    def _mocked_listdir(dummy):
        return [os.path.basename(i) for i in mocked_gpg_files.get_files()]

    monkeypatch.setattr(trustedgpgkeys.os, 'listdir', _mocked_listdir)
    monkeypatch.setattr(trustedgpgkeys, 'get_path_to_gpg_certs', lambda: '/mydir/')
    monkeypatch.setattr(trustedgpgkeys, 'get_gpg_fp_from_file', mocked_gpg_files)

    pubkeys = trustedgpgkeys._get_pubkeys(installed_rpms)
    assert len(pubkeys) == len(rpm_fps + file_fps)
    assert set(rpm_fps) == {pkey.fingerprint for pkey in pubkeys if pkey.rpmdb}
    assert set(file_fps) == {pkey.fingerprint for pkey in pubkeys if not pkey.rpmdb}
    assert list({pkey.filename for pkey in pubkeys if not pkey.rpmdb})[0] == '/mydir/myfile'


def test_process(monkeypatch):
    """
    Executes the "main" function
    """
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        msgs=[_get_test_installed_rmps(['9570ff31'])])
    )
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(api, 'current_logger', logger_mocked())
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    monkeypatch.setattr(trustedgpgkeys, '_get_pubkeys', get_pubkeys_from_rpms)

    trustedgpgkeys.process()
    assert api.produce.called == 1
    assert isinstance(api.produce.model_instances[0], TrustedGpgKeys)
    assert reporting.create_report.called == 0



from collections import defaultdict

from leapp import reporting
from leapp.libraries.common.config import architecture

FMT_LIST_SEPARATOR = '\n    - '
ZIPL_CONFIG_PATH = '/etc/zipl.conf'


def is_rescue_entry(boot_entry):
    """
    Determines whether the given boot entry is rescue.

    :param BootEntry boot_entry: Boot entry to assess
    :return: True is the entry is rescue
    :rtype:  bool
    """
    return 'rescue' in boot_entry.kernel_image.lower()


def inhibit_if_multiple_zipl_rescue_entries_present(bootloader_config):
    """
    Inhibits the upgrade if we are running on s390x and the bootloader configuration
    contains multiple rescue boot entries.

    A boot entry is recognized as a rescue entry when its title contains the `rescue` substring.

    :param SourceBootloaderConfiguration bootloader_config: The configuration of the source boot loader.
    """

    # Keep the whole information about boot entries not just their count as
    # we want to provide user with the details
    rescue_entries = []
    for boot_entry in bootloader_config.entries:
        if is_rescue_entry(boot_entry):
            rescue_entries.append(boot_entry)

    if len(rescue_entries) > 1:
        # Prepare the list of available rescue entries for user
        rescue_entries_text = ''
        for rescue_entry in rescue_entries:
            rescue_entries_text += '{0}{1}'.format(FMT_LIST_SEPARATOR, rescue_entry.title)

        summary = ('The Zipl configuration file {0} contains multiple rescue boot entries preventing migration '
                   'to BLS. Problematic entries: {1}')

        reporting.create_report([
            reporting.Title('Multiple rescue boot entries present in the bootloader configuration.'),
            reporting.Summary(summary.format(ZIPL_CONFIG_PATH, rescue_entries_text)),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.BOOT]),
            reporting.Remediation(hint='Remove rescue boot entries from the configuration and leave just one.'),
            reporting.Groups([reporting.Groups.INHIBITOR])
        ])


def extract_kernel_version(kernel_img_path):
    """
    Extracts the kernel version out of the given image path.

    The extraction logic is designed to closely mimic the logic Zipl configuration to BLS
    conversion script works, so that it is possible to identify the possible issues with kernel
    images.

    :param str kernel_img_path: The path to the kernel image.
    :returns: Extracted kernel version from the given path
    :rtype: str
    """

    # Mimic bash substitution used in the conversion script, see:
    # https://github.com/ibm-s390-linux/s390-tools/blob/b5604850ab66f862850568a37404faa647b5c098/scripts/zipl-switch-to-blscfg#L168
    if 'vmlinuz-' in kernel_img_path:
        fragments = kernel_img_path.rsplit('/vmlinuz-', 1)
        return fragments[1] if len(fragments) > 1 else fragments[0]

    fragments = kernel_img_path.rsplit('/', 1)
    return fragments[1] if len(fragments) > 1 else fragments[0]


def inhibit_if_entries_share_kernel_version(bootloader_config):
    """
    Inhibits the upgrade if there are boot entries sharing the same kernel image version.

    The logic of identification whether the images are the same mimics the zipl-switch-to-blscfg, as it fails
    to perform the conversion if there are entries with the same kernel image.

    :param SourceBootloaderConfiguration bootloader_config: The configuration of the source boot loader.
    """

    used_kernel_versions = defaultdict(list)  # Maps images to the boot entries in which they are used
    for boot_entry in bootloader_config.entries:
        if is_rescue_entry(boot_entry):
            # Rescue entries are handled differently and their images should not cause naming collisions
            continue

        kernel_version = extract_kernel_version(boot_entry.kernel_image)
        used_kernel_versions[kernel_version].append(boot_entry)

    versions_used_multiple_times = []
    for version, version_boot_entries in used_kernel_versions.items():
        if len(version_boot_entries) > 1:
            # Keep the information about entries for the report
            versions_used_multiple_times.append((version, version_boot_entries))

    if versions_used_multiple_times:
        problematic_entries_details = ''
        for version, version_boot_entries in versions_used_multiple_times:
            entry_titles = ['"{0}"'.format(entry.title) for entry in version_boot_entries]
            problematic_entries_details += '{0}{1} (found in entries: {2})'.format(
                    FMT_LIST_SEPARATOR,
                    version,
                    ', '.join(entry_titles)
            )

        summary = ('The zipl configuration file {0} contains boot entries sharing the same kernel version '
                   'preventing migration to BLS. Kernel versions shared: {1}')
        reporting.create_report([
            reporting.Title('Boot entries sharing the same kernel version found.'),
            reporting.Summary(summary.format(ZIPL_CONFIG_PATH, problematic_entries_details)),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.BOOT]),
            reporting.Remediation(
                hint='Remove boot entries sharing the same kernel version from the configuration and leave just one.'),
            reporting.Groups([reporting.Groups.INHIBITOR])
        ])


def inhibit_if_invalid_zipl_configuration(bootloader_config):
    if not architecture.matches_architecture(architecture.ARCH_S390X):
        # Zipl is used only on s390x
        return

    inhibit_if_multiple_zipl_rescue_entries_present(bootloader_config)
    inhibit_if_entries_share_kernel_version(bootloader_config)



from leapp.actors import Actor
from leapp.libraries.stdlib import CalledProcessError, run
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class NetworkManagerUpdateService(Actor):
    """
    Updates NetworkManager services status.

    On Red Hat Enterprise Linux 7 if the NetworkManager service was disabled and
    NetworkManager-wait-online enabled, the former would not be started. This changed on Red Hat
    Enterprise Linux 8, where NM-w-o 'Requires' NM and so NM can be started even if disabled. Upon
    upgrade, to keep the previous behavior we must disable NM-w-o when NM is disabled.

    See also:
      https://bugzilla.redhat.com/show_bug.cgi?id=1520865
    """

    name = 'network_manager_update_service'
    consumes = ()
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        nm_enabled = self.unit_enabled('NetworkManager.service')
        nmwo_enabled = self.unit_enabled('NetworkManager-wait-online.service')
        self.log_services_state('initial', nm_enabled, nmwo_enabled)

        if not nm_enabled and nmwo_enabled:
            self.log.info('Disabling NetworkManager-wait-online.service')

            try:
                run(['systemctl', 'disable', 'NetworkManager-wait-online.service'])
            except (OSError, CalledProcessError) as e:
                self.log.warning('Error disabling NetworkManager-wait-online.service: {}'.format(e))
                return

            nm_enabled = self.unit_enabled('NetworkManager.service')
            nmwo_enabled = self.unit_enabled('NetworkManager-wait-online.service')
            self.log_services_state('after upgrade', nm_enabled, nmwo_enabled)

    def log_services_state(self, detail, nm, nmwo):
        self.log.info('Services state ({}):'.format(detail))
        self.log.info(' - NetworkManager            : {}'.format('enabled' if nm else 'disabled'))
        self.log.info(' - NetworkManager-wait-online: {}'.format('enabled' if nmwo else 'disabled'))

    def unit_enabled(self, name):
        try:
            ret = run(['systemctl', 'is-enabled', name], split=True)['stdout']
            if ret:
                enabled = ret[0] == 'enabled'
            else:
                enabled = False
        except (OSError, CalledProcessError):
            enabled = False
        return enabled



from leapp.actors import Actor
from leapp.libraries.actor import setetcreleasever
from leapp.models import PkgManagerInfo, RHUIInfo
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class SetEtcReleaseVer(Actor):
    """
    Release version in /etc/dnf/vars/releasever will be set to the current target release

    If Leapp detects "releasever" variable is either configured through DNF/YUM configuration
    file and/or the system is using RHUI infrastructure, release version will be set to the target
    release version in order to avoid issues with repofile URLs (when --release option is not provided)
    in cases where there is the previous major.minor version value in the configuration. This will also
    ensure the system stays on the target version after the upgrade.
    """

    name = 'set_etc_releasever'
    consumes = (PkgManagerInfo, RHUIInfo)
    produces = ()
    tags = (IPUWorkflowTag, ApplicationsPhaseTag)

    def process(self):
        setetcreleasever.process()



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class RepositoriesSetupTasks(Model):
    """
    Information about repositories that must be managed in order to complete upgrade process.

    'to_enable' field consists of a list of repositories that should be enabled in order to complete
    upgrade process. This information should be processed by an actor dedicated to manage
    repositories.
    """
    topic = SystemFactsTopic

    to_enable = fields.List(fields.String(), default=[])



#!/usr/bin/python
#
# Convert ntp configuration to chrony
#
# Copyright (C) 2018-2019  Miroslav Lichvar <mlichvar@redhat.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from __future__ import print_function

import argparse
import ipaddress
import logging
import os
import re
import subprocess
import sys

import six

# python2 compatibility hacks
if six.PY2:
    from io import open
    reload(sys)  # pylint: disable=undefined-variable # noqa: F821
    sys.setdefaultencoding("utf-8")


class NtpConfiguration(object):
    def __init__(self, root_dir, ntp_conf, step_tickers):
        self.root_dir = root_dir if root_dir != "/" else ""
        self.ntp_conf_path = ntp_conf
        self.step_tickers_path = step_tickers

        # Read and write files using an 8-bit transparent encoding
        self.file_encoding = "latin-1"
        self.enabled_services = set()
        self.step_tickers = []
        self.time_sources = []
        self.fudges = {}
        self.restrictions = {
                # Built-in defaults
                ipaddress.ip_network(u"0.0.0.0/0"): set(),
                ipaddress.ip_network(u"::/0"): set(),
        }
        self.keyfile = ""
        self.keys = []
        self.trusted_keys = []
        self.driftfile = ""
        self.statistics = []
        self.leapfile = ""
        self.tos_options = {}
        self.ignored_directives = set()
        self.ignored_lines = []

        # self.detect_enabled_services()
        self.parse_step_tickers()
        self.parse_ntp_conf()

    def detect_enabled_services(self):
        for service in ["ntpdate", "ntpd", "ntp-wait"]:
            service_path = os.path.join(self.root_dir,
                                        "etc/systemd/system/multi-user.target.wants/{}.service".format(service))
            if os.path.islink(service_path):
                self.enabled_services.add(service)
        logging.info("Enabled services found in /etc/systemd/system: %s",
                     " ".join(self.enabled_services))

    def parse_step_tickers(self):
        if not self.step_tickers_path:
            return

        path = os.path.join(self.root_dir, self.step_tickers_path)
        if not os.path.isfile(path):
            logging.info("Missing %s", path)
            return

        with open(path, encoding=self.file_encoding) as f:
            for line in f:
                line = line[:line.find('#')]

                words = line.split()

                if not words:
                    continue

                self.step_tickers.extend(words)

    def parse_ntp_conf(self, path=None):
        if path is None:
            path = os.path.join(self.root_dir, self.ntp_conf_path)

        with open(path, encoding=self.file_encoding) as f:
            logging.info("Reading %s", path)

            for line in f:
                line = line[:line.find('#')]

                words = line.split()

                if not words:
                    continue

                if not self.parse_directive(words):
                    self.ignored_lines.append(line)

    def parse_directive(self, words):
        res = True
        name = words.pop(0)
        if name.startswith("logconfig"):
            name = "logconfig"

        if words:
            if name in ["server", "peer", "pool"]:
                res = self.parse_source(name, words)
            elif name == "fudge":
                res = self.parse_fudge(words)
            elif name == "restrict":
                res = self.parse_restrict(words)
            elif name == "tos":
                res = self.parse_tos(words)
            elif name == "includefile":
                res = self.parse_includefile(words)
            elif name == "keys":
                res = self.parse_keys(words)
            elif name == "trustedkey":
                res = self.parse_trustedkey(words)
            elif name == "driftfile":
                self.driftfile = words[0]
            elif name == "statistics":
                self.statistics = words
            elif name == "leapfile":
                self.leapfile = words[0]
            else:
                self.ignored_directives.add(name)
                res = False
        else:
            self.ignored_directives.add(name)
            res = False

        return res

    def parse_source(self, source_type, words):
        ipv4_only = False
        ipv6_only = False
        source = {
                "type": source_type,
                "options": []
        }

        if words[0] == "-4":
            ipv4_only = True
            words.pop(0)
        elif words[0] == "-6":
            ipv6_only = True
            words.pop(0)

        if not words:
            return False

        source["address"] = words.pop(0)

        # Check if -4/-6 corresponds to the address and ignore hostnames
        if ipv4_only or ipv6_only:
            try:
                version = ipaddress.ip_address(source["address"]).version
                if (ipv4_only and version != 4) or (ipv6_only and version != 6):
                    return False
            except ValueError:
                return False

        if source["address"].startswith("127.127."):
            if not source["address"].startswith("127.127.1."):
                # Ignore non-LOCAL refclocks
                return False

        while words:
            if len(words) >= 2 and words[0] in ["minpoll", "maxpoll", "version", "key"]:
                source["options"].append((words[0], words[1]))
                words = words[2:]
            elif words[0] in ["burst", "iburst", "noselect", "prefer", "true", "xleave"]:
                source["options"].append((words[0],))
                words.pop(0)
            else:
                return False

        self.time_sources.append(source)
        return True

    def parse_fudge(self, words):
        address = words.pop(0)
        options = {}

        while words:
            if len(words) >= 2 and words[0] in ["stratum"]:
                if not words[1].isdigit():
                    return False
                options[words[0]] = int(words[1])
                words = words[2:]
            elif len(words) >= 2:
                words = words[2:]
            else:
                return False

        self.fudges[address] = options
        return True

    def parse_restrict(self, words):
        ipv4_only = False
        ipv6_only = False
        flags = set()
        mask = ""

        if words[0] == "-4":
            ipv4_only = True
            words.pop(0)
        elif words[0] == "-6":
            ipv6_only = True
            words.pop(0)

        if not words:
            return False

        address = words.pop(0)

        while words:
            if len(words) >= 2 and words[0] == "mask":
                mask = words[1]
                words = words[2:]
            else:
                if words[0] not in ["kod", "nomodify", "notrap", "nopeer", "noquery",
                                    "limited", "ignore", "noserve"]:
                    return False
                flags.add(words[0])
                words.pop(0)

        # Convert to IP network(s), ignoring restrictions with hostnames
        networks = []
        if address == "default" and not mask:
            if not ipv6_only:
                networks.append(ipaddress.ip_network(u"0.0.0.0/0"))
            if not ipv4_only:
                networks.append(ipaddress.ip_network(u"::/0"))
        else:
            try:
                if mask:
                    # Count bits in the mask (ipaddress does not support
                    # expanded IPv6 netmasks)
                    mask_ip = ipaddress.ip_address(mask)
                    mask_str = "{0:0{1}b}".format(int(mask_ip), mask_ip.max_prefixlen)
                    networks.append(ipaddress.ip_network(
                                    u"{}/{}".format(address, len(mask_str.rstrip('0')))))
                else:
                    networks.append(ipaddress.ip_network(address))
            except ValueError:
                return False

            if (ipv4_only and networks[-1].version != 4) or \
                    (ipv6_only and networks[-1].version != 6):
                return False

        for network in networks:
            self.restrictions[network] = flags

        return True

    def parse_tos(self, words):
        options = {}
        while words:
            if len(words) >= 2 and words[0] in ["minsane", "orphan"]:
                if not words[1].isdigit():
                    return False
                options[words[0]] = int(words[1])
                words = words[2:]
            elif len(words) >= 2 and words[0] in ["maxdist"]:
                # Check if it is a float value
                if not words[1].replace('.', '', 1).isdigit():
                    return False
                options[words[0]] = float(words[1])
                words = words[2:]
            else:
                return False

        self.tos_options.update(options)

        return True

    def parse_includefile(self, words):
        path = os.path.join(self.root_dir, words[0])
        if not os.path.isfile(path):
            return False

        self.parse_ntp_conf(path)
        return True

    def parse_keys(self, words):
        keyfile = words[0]
        path = os.path.join(self.root_dir, keyfile)
        if not os.path.isfile(path):
            logging.info("Missing %s", path)
            return False

        with open(path, encoding=self.file_encoding) as f:
            logging.info("Reading %s", path)
            keys = []
            for line in f:
                words = line.split()
                if len(words) < 3 or not words[0].isdigit():
                    continue
                keys.append((int(words[0]), words[1], words[2]))

            self.keyfile = keyfile
            self.keys = keys

        return True

    def parse_trustedkey(self, words):
        key_ranges = []
        for word in words:
            if word.isdigit():
                key_ranges.append((int(word), int(word)))
            elif re.match("^[0-9]+-[0-9]+$", word):
                first, last = word.split("-")
                key_ranges.append((int(first), int(last)))
            else:
                return False

        self.trusted_keys = key_ranges
        return True

    def write_chrony_configuration(self, chrony_conf_path, chrony_keys_path,
                                   dry_run=False, backup=False):
        chrony_conf = self.get_chrony_conf(chrony_keys_path)
        logging.debug("Generated %s:\n%s", chrony_conf_path, chrony_conf)

        if not dry_run:
            self.write_file(chrony_conf_path, 0o644, chrony_conf, backup)

        chrony_keys = self.get_chrony_keys()
        if chrony_keys:
            logging.debug("Generated %s:\n%s", chrony_keys_path, chrony_keys)

        if not dry_run:
            self.write_file(chrony_keys_path, 0o640, chrony_keys, backup)

    def get_processed_time_sources(self):
        # Convert {0,1,2,3}.*pool.ntp.org servers to 2.*pool.ntp.org pools

        # Make shallow copies of all sources (only type will be modified)
        time_sources = [s.copy() for s in self.time_sources]

        pools = {}
        for source in time_sources:
            if source["type"] != "server":
                continue
            m = re.match("^([0123])(\\.\\w+)?\\.pool\\.ntp\\.org$", source["address"])
            if m is None:
                continue
            number = m.group(1)
            zone = m.group(2)
            if zone not in pools:
                pools[zone] = []
            pools[zone].append((int(number), source))

        remove_servers = set()
        for zone, pool in pools.items():
            # sort and skip all pools not in [0, 3] range
            pool.sort()
            if [num for num, source in pool] != [0, 1, 2, 3]:
                # only exact group of 4 servers can be converted, nothing to do here
                continue
            # verify that parameters are the same for all servers in the pool
            if not all([p[1]["options"] == pool[0][1]["options"] for p in pool]):
                break
            remove_servers.update([pool[i][1]["address"] for i in [0, 1, 3]])
            pool[2][1]["type"] = "pool"

        processed_sources = []
        for source in time_sources:
            if source["type"] == "server" and source["address"] in remove_servers:
                continue
            processed_sources.append(source)
        return processed_sources

    def get_chrony_conf_sources(self):
        conf = ""

        if self.step_tickers:
            conf += "# Specify NTP servers used for initial correction.\n"
            conf += "initstepslew 0.1 {}\n".format(" ".join(self.step_tickers))
            conf += "\n"

        conf += "# Specify time sources.\n"

        for source in self.get_processed_time_sources():
            address = source["address"]
            if address.startswith("127.127."):
                if address.startswith("127.127.1."):
                    continue
                # No other refclocks are expected from the parser
                assert False
            else:
                conf += "{} {}".format(source["type"], address)
                for option in source["options"]:
                    if option[0] in ["minpoll", "maxpoll", "version", "key",
                                     "iburst", "noselect", "prefer", "xleave"]:
                        conf += " {}".format(" ".join(option))
                    elif option[0] == "burst":
                        conf += " presend 6"
                    elif option[0] == "true":
                        conf += " trust"
                    else:
                        # No other options are expected from the parser
                        assert False
                conf += "\n"
        conf += "\n"

        return conf

    def get_chrony_conf_allows(self):
        allowed_networks = [n for n in self.restrictions
                            if "ignore" not in self.restrictions[n] and "noserve" not in self.restrictions[n]]

        conf = ""
        for network in sorted(allowed_networks, key=lambda n: (n.version, n)):
            if network.num_addresses > 1:
                conf += "allow {}\n".format(network)
            else:
                conf += "allow {}\n".format(network.network_address)

        if conf:
            conf = "# Allow NTP client access.\n" + conf
            conf += "\n"

        return conf

    def get_chrony_conf_cmdallows(self):

        def _is_network_allowed(network):
            return ("ignore" not in self.restrictions[network] and "noquery" not in self.restrictions[network] and
                    network != ipaddress.ip_network(u"127.0.0.1/32") and network != ipaddress.ip_network(u"::1/128"))

        allowed_networks = [n for n in self.restrictions if _is_network_allowed(n)]

        ip_versions = set()
        conf = ""
        for network in sorted(allowed_networks, key=lambda n: (n.version, n)):
            ip_versions.add(network.version)
            if network.num_addresses > 1:
                conf += "cmdallow {}\n".format(network)
            else:
                conf += "cmdallow {}\n".format(network.network_address)

        if conf:
            conf = "# Allow remote monitoring.\n" + conf
            if 4 in ip_versions:
                conf += "bindcmdaddress 0.0.0.0\n"
            if 6 in ip_versions:
                conf += "bindcmdaddress ::\n"
            conf += "\n"

        return conf

    def get_chrony_conf(self, chrony_keys_path):
        local_stratum = 0
        maxdistance = 0.0
        minsources = 1
        orphan_stratum = 0
        logs = []

        for source in self.time_sources:
            address = source["address"]
            if address.startswith("127.127.1."):
                if address in self.fudges and "stratum" in self.fudges[address]:
                    local_stratum = self.fudges[address]["stratum"]
                else:
                    local_stratum = 5

        if "maxdist" in self.tos_options:
            maxdistance = self.tos_options["maxdist"]
        if "minsane" in self.tos_options:
            minsources = self.tos_options["minsane"]
        if "orphan" in self.tos_options:
            orphan_stratum = self.tos_options["orphan"]

        if "clockstats" in self.statistics:
            logs.append("refclocks")
        if "loopstats" in self.statistics:
            logs.append("tracking")
        if "peerstats" in self.statistics:
            logs.append("statistics")
        if "rawstats" in self.statistics:
            logs.append("measurements")

        conf = "# This file was converted from {}{}.\n".format(
                    self.ntp_conf_path,
                    " and " + self.step_tickers_path if self.step_tickers_path else "")
        conf += "\n"

        if self.ignored_lines:
            conf += "# The following directives were ignored in the conversion:\n"

            for line in self.ignored_lines:
                # Remove sensitive information
                line = re.sub(r"\s+pw\s+\S+", " pw XXX", line.rstrip())
                conf += "# " + line + "\n"
            conf += "\n"

        conf += self.get_chrony_conf_sources()

        conf += "# Record the rate at which the system clock gains/losses time.\n"
        if not self.driftfile:
            conf += "#"
        conf += "driftfile /var/lib/chrony/drift\n"
        conf += "\n"

        conf += "# Allow the system clock to be stepped in the first three updates\n"
        conf += "# if its offset is larger than 1 second.\n"
        conf += "makestep 1.0 3\n"
        conf += "\n"

        conf += "# Enable kernel synchronization of the real-time clock (RTC).\n"
        conf += "rtcsync\n"
        conf += "\n"

        conf += "# Enable hardware timestamping on all interfaces that support it.\n"
        conf += "#hwtimestamp *\n"
        conf += "\n"

        if maxdistance > 0.0:
            conf += "# Specify the maximum distance of sources to be selectable.\n"
            conf += "maxdistance {}\n".format(maxdistance)
            conf += "\n"

        conf += "# Increase the minimum number of selectable sources required to adjust\n"
        conf += "# the system clock.\n"
        if minsources > 1:
            conf += "minsources {}\n".format(minsources)
        else:
            conf += "#minsources 2\n"
        conf += "\n"

        conf += self.get_chrony_conf_allows()

        conf += self.get_chrony_conf_cmdallows()

        conf += "# Serve time even if not synchronized to a time source.\n"
        if 0 < orphan_stratum < 16:
            conf += "local stratum {} orphan\n".format(orphan_stratum)
        elif 0 < local_stratum < 16:
            conf += "local stratum {}\n".format(local_stratum)
        else:
            conf += "#local stratum 10\n"
        conf += "\n"

        conf += "# Specify file containing keys for NTP authentication.\n"
        conf += "keyfile {}\n".format(chrony_keys_path)
        conf += "\n"

        conf += "# Get TAI-UTC offset and leap seconds from the system tz database.\n"
        conf += "leapsectz right/UTC\n"
        conf += "\n"

        conf += "# Specify directory for log files.\n"
        conf += "logdir /var/log/chrony\n"
        conf += "\n"

        conf += "# Select which information is logged.\n"
        if logs:
            conf += "log {}\n".format(" ".join(logs))
        else:
            conf += "#log measurements statistics tracking\n"

        return conf

    def get_chrony_keys(self):
        if not self.keyfile:
            return ""

        keys = "# This file was converted from {}.\n".format(self.keyfile)
        keys += "\n"

        for key in self.keys:
            key_id = key[0]
            key_type = key[1]
            password = key[2]

            if key_type in ["m", "M"]:
                key_type = "MD5"
            elif key_type not in ["MD5", "SHA1", "SHA256", "SHA384", "SHA512"]:
                continue

            prefix = "ASCII" if len(password) <= 20 else "HEX"

            for first, last in self.trusted_keys:
                if first <= key_id <= last:
                    trusted = True
                    break
            else:
                trusted = False

            # Disable keys that were not marked as trusted
            if not trusted:
                keys += "#"

            keys += "{key_id} {key_type} {prefix}:{password}\n".format(
                    key_id=key_id, key_type=key_type, prefix=prefix, password=password)

        return keys

    def write_file(self, path, mode, content, backup):
        path = self.root_dir + path
        if backup and os.path.isfile(path):
            os.rename(path, path + ".old")

        with open(os.open(path, os.O_CREAT | os.O_WRONLY | os.O_EXCL, mode), "w",
                  encoding=self.file_encoding) as f:
            logging.info("Writing %s", path)
            f.write(u"" + content)

        # Fix SELinux context if restorecon is installed
        try:
            subprocess.call(["restorecon", path])
        except OSError:
            pass


def main():
    parser = argparse.ArgumentParser(description="Convert ntp configuration to chrony.")
    parser.add_argument("-r", "--root", dest="roots", default=["/"], nargs="+",
                        metavar="DIR", help="specify root directory (default /)")
    parser.add_argument("--ntp-conf", action="store", default="/etc/ntp.conf",
                        metavar="FILE", help="specify ntp config (default /etc/ntp.conf)")
    parser.add_argument("--step-tickers", action="store", default="",
                        metavar="FILE", help="specify ntpdate step-tickers config (no default)")
    parser.add_argument("--chrony-conf", action="store", default="/etc/chrony.conf",
                        metavar="FILE", help="specify chrony config (default /etc/chrony.conf)")
    parser.add_argument("--chrony-keys", action="store", default="/etc/chrony.keys",
                        metavar="FILE", help="specify chrony keyfile (default /etc/chrony.keys)")
    parser.add_argument("-b", "--backup", action="store_true", help="backup existing configs before writing")
    parser.add_argument("-L", "--ignored-lines", action="store_true", help="print ignored lines")
    parser.add_argument("-D", "--ignored-directives", action="store_true",
                        help="print names of ignored directives")
    parser.add_argument("-n", "--dry-run", action="store_true", help="don't make any changes")
    parser.add_argument("-v", "--verbose", action="count", default=0, help="increase verbosity")

    args = parser.parse_args()

    logging.basicConfig(format="%(message)s",
                        level=[logging.ERROR, logging.INFO, logging.DEBUG][min(args.verbose, 2)])

    for root in args.roots:
        conf = NtpConfiguration(root, args.ntp_conf, args.step_tickers)

        if args.ignored_lines:
            for line in conf.ignored_lines:
                print(line)

        if args.ignored_directives:
            for directive in conf.ignored_directives:
                print(directive)

        conf.write_chrony_configuration(args.chrony_conf, args.chrony_keys, args.dry_run, args.backup)


if __name__ == "__main__":
    main()



import shutil

from leapp.actors import Actor
from leapp.libraries.common import dnfplugin
from leapp.libraries.stdlib import run
from leapp.models import (
    DNFPluginTask,
    DNFWorkaround,
    FilteredRpmTransactionTasks,
    RHSMInfo,
    StorageInfo,
    TargetUserSpaceInfo,
    TransactionCompleted,
    UsedTargetRepositories,
    XFSPresence
)
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class DnfUpgradeTransaction(Actor):
    """
    Setup and call DNF upgrade command

    Based on previously calculated RPM transaction data, this actor will setup and call
    rhel-upgrade DNF plugin with necessary parameters
    """

    name = 'dnf_upgrade_transaction'
    consumes = (
        DNFPluginTask,
        DNFWorkaround,
        FilteredRpmTransactionTasks,
        RHSMInfo,
        StorageInfo,
        TargetUserSpaceInfo,
        UsedTargetRepositories,
        XFSPresence
    )
    produces = (TransactionCompleted,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        src_rhsm_info = next(self.consume(RHSMInfo), None)
        if src_rhsm_info:
            for prod_cert in src_rhsm_info.existing_product_certificates:
                run(['rm', '-f', prod_cert])

        used_repos = self.consume(UsedTargetRepositories)
        storage_info = next(self.consume(StorageInfo), None)
        plugin_info = list(self.consume(DNFPluginTask))
        tasks = next(self.consume(FilteredRpmTransactionTasks), FilteredRpmTransactionTasks())
        target_userspace_info = next(self.consume(TargetUserSpaceInfo), None)
        xfs_info = next(self.consume(XFSPresence), XFSPresence())

        dnfplugin.perform_transaction_install(
            tasks=tasks, used_repos=used_repos, storage_info=storage_info, target_userspace_info=target_userspace_info,
            plugin_info=plugin_info, xfs_info=xfs_info
        )
        self.produce(TransactionCompleted())
        userspace = next(self.consume(TargetUserSpaceInfo), None)
        if userspace:
            try:
                shutil.rmtree(userspace.path)
            except EnvironmentError:
                self.log.info("Failed to remove temporary userspace - error ignored", exc_info=True)



from leapp.models import FirewalldUsedObjectNames
from leapp.reporting import Report


def test_actor_firewalldcheckservicetftpclient(current_actor_context):
    services = ['cockpit', 'tftp-client', 'ssh', 'https']
    policies = []
    zones = []

    current_actor_context.feed(FirewalldUsedObjectNames(services=services,
                                                        policies=policies,
                                                        zones=zones))

    current_actor_context.run()
    assert current_actor_context.consume(Report)


def test_actor_firewalldcheckservicetftpclient_negative(current_actor_context):
    services = ['cockpit', 'ssh', 'https']
    policies = []
    zones = []

    current_actor_context.feed(FirewalldUsedObjectNames(services=services,
                                                        policies=policies,
                                                        zones=zones))

    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_firewalldcheckservicetftpclient_negative2(current_actor_context):
    current_actor_context.feed(FirewalldUsedObjectNames(services=[],
                                                        policies=[],
                                                        zones=[]))

    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_firewalldcheckservicetftpclient_negative3(current_actor_context):
    current_actor_context.feed(FirewalldUsedObjectNames())

    current_actor_context.run()
    assert not current_actor_context.consume(Report)



import os

from leapp.libraries.actor import multipathconfupdate
from leapp.libraries.common import multipathutil
from leapp.models import MultipathConfFacts8to9, MultipathConfig8to9

BEFORE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files/before')
AFTER_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files/after')

converted_data = {}


def build_config(pathname, config_dir, enable_foreign_exists, invalid_regexes_exist, allow_usb_exists):
    return MultipathConfig8to9(
        pathname=pathname,
        config_dir=config_dir,
        enable_foreign_exists=enable_foreign_exists,
        invalid_regexes_exist=invalid_regexes_exist,
        allow_usb_exists=allow_usb_exists,
    )


def build_facts(confs):
    return MultipathConfFacts8to9(configs=confs)


def mock_read_config(path):
    """convert to full pathname"""
    return multipathutil.read_config_orig(os.path.join(BEFORE_DIR, path))


def mock_write_config(path, contents):
    converted_data[path] = contents


default_rhel8_conf = build_config(
    'default_rhel8.conf', None, True, False, False)

all_the_things_conf = build_config(
    'all_the_things.conf', None, False, True, False)

converted_the_things_conf = build_config(
    'converted_the_things.conf', None, True, False, True)

idempotent_conf = build_config(
    'converted_the_things.conf', None, False, True, False)

complicated_conf = build_config(
    'complicated.conf', '/etc/multipath/conf.d', True, True, False)

no_foreign_conf = build_config(
    'no_foreign.conf', None, False, True, True)

allow_usb_conf = build_config(
    'allow_usb.conf', None, False, False, True)

no_defaults_conf = build_config(
    'no_defaults.conf', None, False, True, False)

two_defaults_conf = build_config(
    'two_defaults.conf', None, True, False, False)

empty_conf = build_config(
    'empty.conf', None, False, False, False)

missing_dir_conf = build_config(
    'missing_dir.conf', 'missing', False, True, False)

not_set_dir_conf = build_config(
    'not_set_dir.conf', 'conf1.d', False, True, False)

empty1_conf = build_config(
    'conf1.d/empty.conf', None, False, False, False)

nothing_important_conf = build_config(
    'conf1.d/nothing_important.conf', 'this_gets_ignored', False, False, False)

set_in_dir_conf = build_config(
    'set_in_dir.conf', 'conf2.d', False, False, False)

all_true_conf = build_config(
    'conf2.d/all_true.conf', None, True, True, True)

empty_dir_conf = build_config(
    'empty_dir.conf', 'conf3.d', False, False, False)

facts_list = [build_facts([default_rhel8_conf]),
              build_facts([all_the_things_conf]),
              build_facts([converted_the_things_conf]),
              build_facts([complicated_conf]),
              build_facts([no_foreign_conf]),
              build_facts([allow_usb_conf]),
              build_facts([no_defaults_conf]),
              build_facts([two_defaults_conf]),
              build_facts([empty_conf]),
              build_facts([missing_dir_conf]),
              build_facts([empty_dir_conf]),
              build_facts([not_set_dir_conf, empty1_conf, nothing_important_conf]),
              build_facts([set_in_dir_conf, all_true_conf]),
              build_facts([idempotent_conf])]


def _test_facts(facts):
    multipathconfupdate.update_configs(facts)
    for config in facts.configs:
        expected_data = multipathutil.read_config_orig(os.path.join(AFTER_DIR, config.pathname))
        if config.pathname in converted_data:
            assert converted_data[config.pathname] == expected_data
        else:
            assert expected_data is None


def test_all_facts(monkeypatch):
    monkeypatch.setattr(multipathutil, 'read_config_orig', multipathutil.read_config, raising=False)
    monkeypatch.setattr(multipathutil, 'read_config', mock_read_config)
    monkeypatch.setattr(multipathutil, 'write_config', mock_write_config)
    for facts in facts_list:
        _test_facts(facts)
        converted_data.clear()



import pytest

from leapp.libraries.actor.cupsscanner import include_directive_check

testdata = (
    (
        {
            '/etc/cups/cupsd.conf': '\n'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf'],
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'Include smth.conf\n',
            'smth.conf': '\n'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf', 'smth.conf'],
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'Include smth.conf\nInclude smb.conf\n',
            'smth.conf': '\n',
            'smb.conf': '\n'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf', 'smth.conf',
                               'smb.conf'],
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'Include smth.conf\n',
            'smth.conf': 'Include smb.conf\n',
            'smb.conf': '\n'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf', 'smth.conf',
                               'smb.conf'],
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'Include smth.conf\n',
            'smth.conf': 'Include smb.conf\n',
            'smb.conf': 'Include any.conf\n',
            'any.conf': '\n'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf', 'smth.conf',
                               'smb.conf', 'any.conf'],
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': '#Include smth.conf'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf']
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': 'Include\n'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf'],
        }
    ),
    (
        {
            '/etc/cups/cupsd.conf': '    Include smth.conf smth_more\n',
            'smth.conf': '\n'
        },
        {
            'included_files': ['/etc/cups/cupsd.conf', 'smth.conf']
        }
    )
)


class MockFileSystem(object):
    def __init__(self, infiles):
        self.files = infiles

    def read(self, path):
        if path in self.files.keys():
            return self.files[path].splitlines(True)
        raise IOError('Error during reading file.')


@pytest.mark.parametrize("files,expected", testdata)
def test_include_directive_check(files, expected):
    f = MockFileSystem(files)

    included_files, error_list = include_directive_check(read_func=f.read)

    assert included_files == expected.get('included_files', [])
    assert error_list == expected.get('error_list', [])



from leapp.actors import Actor
from leapp.libraries.actor import check_consumed_assets
from leapp.models import ConsumedDataAsset, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckConsumedAssets(Actor):
    """
    Check whether Leapp is using correct data assets.
    """

    name = 'check_consumed_assets'
    consumes = (ConsumedDataAsset,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        check_consumed_assets.inhibit_if_assets_with_incorrect_version()



import pytest

from leapp.libraries.actor.cupsmigrate import migrate_certkey


class MockFileSystem(object):
    def __init__(self,
                 infiles):
        self.files = {}
        self.ssl_dir = []
        self.files = infiles
        for path in infiles.keys():
            if path.startswith('/etc/cups/ssl'):
                self.ssl_dir.append(path)

    def readlines(self, path):
        if path in self.files.keys():
            return self.files[path].splitlines(True)
        raise IOError('Error when reading file {} - file '
                      'does not exist.'.format(path))

    def write(self, path, mode, content):
        if isinstance(content, list):
            content = ''.join(content)

        if mode == 'w':
            self.files[path] = content
        else:
            self.files[path] += content

    def copy_to_ssl(self, oldpath):
        self.ssl_dir.append('/etc/cups/ssl/' + oldpath.rsplit('/', 1)[1])


testdata = (
    (
        {
            '/etc/cups/cups-files.conf': 'ifdfdfgfg\n'
        },
        {
            '/etc/cups/cups-files.conf': 'ifdfdfgfg\n'
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerKey /etc/cups/ssl/ser'
                                         'ver.key\nHello world\n',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.key']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': '#ServerKey /etc/cups/ssl/se'
                                         'rver.key\nHello world\n',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': '#ServerKey /etc/cups/ssl/se'
                                         'rver.key\nHello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.key']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate /etc/cups'
                                         '/ssl/server.cert\nHello world\n',
            '/etc/cups/ssl/server.cert': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.cert']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate /etc/cups'
                                         '/ssl/server.cert\nServerKey'
                                         ' /etc/cups/ssl/server.key\n'
                                         'Hello world\n',
            '/etc/cups/ssl/server.cert': '',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.cert',
                        '/etc/cups/ssl/server.key']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate ssl/serve'
                                         'r.cert\nServerKey ssl/serve'
                                         'r.key\nHello world\n',
            '/etc/cups/ssl/server.cert': '',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.cert',
                        '/etc/cups/ssl/server.key']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate ssl/serve'
                                         'r.cert\nServerKey /etc/cups'
                                         '/ssl/server.key\nHello worl'
                                         'd\n',
            '/etc/cups/ssl/server.cert': '',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.cert',
                        '/etc/cups/ssl/server.key']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate ssl/serve'
                                         'r.cert\nServerKey /somewher'
                                         'e/else/server.key\nHello wo'
                                         'rld\n',
            '/etc/cups/ssl/server.cert': '',
            '/somewhere/else/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.cert',
                        '/etc/cups/ssl/server.key']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate /somewher'
                                         'e/else/server.cert\nServerK'
                                         'ey /etc/cups/ssl/server.key'
                                         '\nHello world\n',
            '/somewhere/else/server.cert': '',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.key',
                        '/etc/cups/ssl/server.cert']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate /somewher'
                                         'e/else/server.cert\nServerK'
                                         'ey /somewhere/else/server.c'
                                         'ert\nHello world\n',
            '/somewhere/else/server.cert': '',
            '/somewhere/else/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n\n# added by L'
                                         'eapp\nServerKeychain /somew'
                                         'here/else\n'
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate /somewher'
                                         'e/else/server.cert\nServerK'
                                         'ey /anywhere/else/server.ke'
                                         'y\nHello world\n',
            '/somewhere/else/server.cert': '',
            '/anywhere/else/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/server.cert',
                        '/etc/cups/ssl/server.key']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate ssl/somedir/'
                                         'server.cert\nHello world\nServ'
                                         'erKey ssl/server.key\n',
            '/etc/cups/ssl/somedir/server.cert': '',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/somedir/server.cert',
                        '/etc/cups/ssl/server.key',
                        '/etc/cups/ssl/server.cert']
        }
    ),
    (
        {
            '/etc/cups/cups-files.conf': 'ServerCertificate /etc/cups/ss'
                                         'l/somedir/server.cert\nHello w'
                                         'orld\nServerKey ssl/server.key\n',
            '/etc/cups/ssl/somedir/server.cert': '',
            '/etc/cups/ssl/server.key': ''
        },
        {
            '/etc/cups/cups-files.conf': 'Hello world\n',
            'ssl-dir': ['/etc/cups/ssl/somedir/server.cert',
                        '/etc/cups/ssl/server.key',
                        '/etc/cups/ssl/server.cert']
        }
    )
)


@pytest.mark.parametrize('files,expected', testdata)
def test_migrate_certkey(files, expected):
    op = MockFileSystem(infiles=files)

    migrate_certkey(op)

    assert op.files.get('/etc/cups/cups-files.conf', None) == expected.get('/etc/cups/cups-files.conf', None)
    assert op.ssl_dir == expected.get('ssl-dir', [])



import pytest

from leapp import reporting
from leapp.libraries.actor.baculacheck import report_installed_packages
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM, RPM


def _generate_rpm_with_name(name):
    """
    Generate new RPM model item with given name.

    Parameters:
        name (str): rpm name

    Returns:
        rpm  (RPM): new RPM object with name parameter set
    """
    return RPM(name=name,
               version='0.1',
               release='1.sm01',
               epoch='1',
               pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51',
               packager='Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>',
               arch='noarch')


@pytest.mark.parametrize('has_director', [
    (True),  # with director
    (False),  # without director
])
def test_actor_execution(monkeypatch, has_director):
    """
    Parametrized helper function for test_actor_* functions.

    First generate list of RPM models based on set arguments. Then, run
    the actor fed with our RPM list. Finally, assert Reports
    according to set arguments.

    Parameters:
        has_director  (bool): bacula-director installed
    """

    # Couple of random packages
    rpms = [_generate_rpm_with_name('sed'),
            _generate_rpm_with_name('htop')]

    if has_director:
        # Add bacula-director
        rpms += [_generate_rpm_with_name('bacula-director')]

    curr_actor_mocked = CurrentActorMocked(msgs=[DistributionSignedRPM(items=rpms)])
    monkeypatch.setattr(api, 'current_actor', curr_actor_mocked)
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    # Executed actor fed with out fake RPMs
    report_installed_packages(_context=api)

    if has_director:
        # Assert for bacula-director package installed
        assert reporting.create_report.called == 1
    else:
        # Assert for no bacula packages installed
        assert not reporting.create_report.called



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class MemoryInfo(Model):
    """
    Represents information about available memory (KiB)
    """

    topic = SystemFactsTopic

    mem_total = fields.Nullable(fields.Integer())
    """
    Total memory in KiB
    """



import leapp.libraries.common.tcpwrappersutils as lib
from leapp.models import DaemonList, TcpWrappersFacts


def test_config_applies_to_daemon_simple():
    daemon_list = DaemonList(value=['vsftpd'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])

    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True
    assert lib.config_applies_to_daemon(facts, 'VsfTpd') is True
    assert lib.config_applies_to_daemon(facts, 'ftp') is False
    assert lib.config_applies_to_daemon(facts, 'foo') is False


def test_config_applies_to_daemon_multiple_lists():
    list1 = DaemonList(value=['vsftpd', 'sendmail'])
    list2 = DaemonList(value=['postfix'])
    facts = TcpWrappersFacts(daemon_lists=[list1, list2])

    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True
    assert lib.config_applies_to_daemon(facts, 'sendmail') is True
    assert lib.config_applies_to_daemon(facts, 'postfix') is True
    assert lib.config_applies_to_daemon(facts, 'foo') is False


def test_config_applies_to_daemon_except():
    list1 = DaemonList(value=['all', 'except', 'sendmail'])
    list2 = DaemonList(value=['postfix'])
    facts = TcpWrappersFacts(daemon_lists=[list1, list2])

    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True
    assert lib.config_applies_to_daemon(facts, 'sendmail') is False
    assert lib.config_applies_to_daemon(facts, 'postfix') is True
    assert lib.config_applies_to_daemon(facts, 'foo') is True

    list1 = DaemonList(value=['all', 'except', 'b*', 'EXCEPT', 'bar'])
    facts = TcpWrappersFacts(daemon_lists=[list1])
    assert lib.config_applies_to_daemon(facts, 'foo') is True
    assert lib.config_applies_to_daemon(facts, 'bar') is True
    assert lib.config_applies_to_daemon(facts, 'baar') is False

    list1 = DaemonList(value=['all', 'except', 'vsftpd'])
    facts = TcpWrappersFacts(daemon_lists=[list1])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False

    list1 = DaemonList(value=['all', 'except', 'all', 'except', 'vsftpd'])
    facts = TcpWrappersFacts(daemon_lists=[list1])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    list1 = DaemonList(value=['all', 'except', 'all', 'except', 'all', 'except', 'vsftpd'])
    facts = TcpWrappersFacts(daemon_lists=[list1])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False


def test_config_applies_to_daemon_except_empty():
    daemon_list = DaemonList(value=['all', 'except'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True


def test_config_applies_to_daemon_with_host():
    list1 = DaemonList(value=['vsftpd@localhost', 'sendmail'])
    list2 = DaemonList(value=['postfix'])
    facts = TcpWrappersFacts(daemon_lists=[list1, list2])

    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True
    assert lib.config_applies_to_daemon(facts, 'sendmail') is True
    assert lib.config_applies_to_daemon(facts, 'postfix') is True
    assert lib.config_applies_to_daemon(facts, 'foo') is False


def test_config_applies_to_daemon_with_host_except():
    daemon_list = DaemonList(value=['vsftpd@localhost', 'except', 'vsftpd'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False

    # It works like this for simplicity.
    daemon_list = DaemonList(value=['vsftpd@localhost', 'except', 'vsftpd@localhost'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['vsftpd@localhost'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['all', 'except', 'vsftpd@localhost'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['all', 'except', 'all', 'except', 'vsftpd@localhost'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['all', 'except', 'all', 'except', 'all',
                                    'except', 'vsftpd@localhost'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['all', 'except', 'all', 'except', 'all', 'except', 'all',
                                    'except', 'vsftpd@localhost'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True


def test_config_applies_to_daemon_empty():
    daemon_list = DaemonList(value=[''])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False

    daemon_list = DaemonList(value=[])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False


def test_config_applies_to_daemon_whole_word():
    daemon_list = DaemonList(value=['ftp'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False


def test_config_applies_to_daemon_asterisk_wildcard():
    daemon_list = DaemonList(value=['*ftp*'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['************'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['*'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['*foo*'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False


def test_config_applies_to_daemon_question_mark_wildcard():
    daemon_list = DaemonList(value=['vs?tpd'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['vsf?tpd'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False

    daemon_list = DaemonList(value=['?'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False

    daemon_list = DaemonList(value=['??????'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True


def test_config_applies_to_daemon_all_wildcard():
    daemon_list = DaemonList(value=['all'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['aLl'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is True

    daemon_list = DaemonList(value=['al'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False

    daemon_list = DaemonList(value=['ll'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False

    daemon_list = DaemonList(value=['valld'])
    facts = TcpWrappersFacts(daemon_lists=[daemon_list])
    assert lib.config_applies_to_daemon(facts, 'vsftpd') is False



from leapp.actors import Actor
from leapp.libraries.actor import checkinsightsautoregister
from leapp.models import InstalledRPM, RpmTransactionTasks
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckInsightsAutoregister(Actor):
    """
    Checks if system can be automatically registered into Red Hat Insights

    The registration is skipped if NO_INSIGHTS_REGISTER=1 environment variable
    is set, the --no-insights-register command line argument present. if the
    system isn't registered with subscription-manager.

    Additionally, the insights-client package is required. If it's missing an
    RpmTransactionTasks is produced to install it during the upgrade.

    A report is produced informing about the automatic registration and
    eventual insights-client package installation.
    """

    name = 'check_insights_auto_register'
    consumes = (InstalledRPM,)
    produces = (Report, RpmTransactionTasks)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        checkinsightsautoregister.process()



import os

import distro
import pytest


@pytest.mark.skipif(os.getuid() != 0, reason='User is not a root')
@pytest.mark.skipif(
    distro.linux_distribution()[0] == 'Fedora',
    reason='default.target.wants does not exists on Fedora distro',
)
def test_create_resume_service(current_actor_context):

    current_actor_context.run()

    service_name = 'leapp_resume.service'
    service_path = '/etc/systemd/system/{}'.format(service_name)
    symlink_path = '/etc/systemd/system/default.target.wants/{}'.format(service_name)

    try:
        assert os.path.isfile(service_path)
        assert os.path.isfile(symlink_path)
    finally:
        os.unlink(service_path)
        os.unlink(symlink_path)



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api
from leapp.models import MemoryInfo


def _get_memoryinfo(filename='/proc/meminfo'):
    """ Returns dict of all memory information from /proc/meminfo file """

    try:
        with open(filename) as fp:
            # format of the lines: Key:  value unit
            # e.g.:                MemTotal:   1024 kB
            return dict((i.split()[0].rstrip(':'), int(i.split()[1])) for i in fp.readlines())
    except IOError:
        raise StopActorExecutionError(
            'Could not read memory values',
            details={'details': 'Error opening file {}'.format(filename)},
        )


def process():
    mem_info = _get_memoryinfo()
    if mem_info:
        api.produce(MemoryInfo(mem_total=mem_info['MemTotal']))



from leapp import reporting
from leapp.actors import Actor
from leapp.models import SelinuxRelabelDecision
from leapp.reporting import create_report, Report
from leapp.tags import FinalizationPhaseTag, IPUWorkflowTag

COMMON_REPORT_TAGS = [reporting.Groups.SELINUX]

related = [reporting.RelatedResource('file', '/.autorelabel')]


class ScheduleSeLinuxRelabeling(Actor):
    """
    Schedule SELinux relabelling.

    If SELinux status was set to permissive or enforcing, a relabelling is necessary.
    """

    name = 'schedule_se_linux_relabelling'
    consumes = (SelinuxRelabelDecision,)
    produces = (Report,)
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):
        for decision in self.consume(SelinuxRelabelDecision):
            if decision.set_relabel:
                try:
                    with open('/.autorelabel', 'w'):
                        pass
                    create_report([
                        reporting.Title('SElinux scheduled for relabelling'),
                        reporting.Summary(
                            '/.autorelabel file touched on root in order to schedule SElinux relabelling.'),
                        reporting.Severity(reporting.Severity.INFO),
                        reporting.Groups(COMMON_REPORT_TAGS),
                    ] + related)

                except EnvironmentError as e:
                    # FIXME: add an "action required" flag later
                    create_report([
                        reporting.Title('Could not schedule SElinux for relabelling'),
                        reporting.Summary('/.autorelabel file could not be created: {}.'.format(e)),
                        reporting.Severity(reporting.Severity.HIGH),
                        reporting.Groups(COMMON_REPORT_TAGS),
                        reporting.Remediation(
                            hint='Please set autorelabelling manually after the upgrade.'
                        ),
                        reporting.Groups([reporting.Groups.FAILURE])
                    ] + related)
                    self.log.critical('Could not schedule SElinux for relabelling: %s.' % e)



from leapp.actors import Actor
from leapp.libraries.common import persistentnetnames
from leapp.models import PersistentNetNamesFactsInitramfs
from leapp.tags import IPUWorkflowTag, LateTestsPhaseTag


class PersistentNetNamesInitramfs(Actor):
    """
    Get network interface information for physical ethernet interfaces with the new kernel in initramfs.

    This actor does exactly the same job as PersistentNetNames actor except that it runs in a later phase.
    """

    name = 'persistentnetnamesinitramfs'
    consumes = ()
    produces = (PersistentNetNamesFactsInitramfs, )
    tags = (LateTestsPhaseTag, IPUWorkflowTag)

    def process(self):
        self.produce(PersistentNetNamesFactsInitramfs(interfaces=list(persistentnetnames.interfaces())))



from leapp import reporting
from leapp.libraries.common.config.version import get_source_major_version

MIGRATION_GUIDE_7 = "https://red.ht/IdM-upgrading-RHEL-7-to-RHEL-8"

# TBD: update the doc url when migration guide 8->9 becomes available
MIGRATION_GUIDE_8 = "https://red.ht/IdM-upgrading-RHEL-8-to-RHEL-9"
MIGRATION_GUIDES = {
    '7': MIGRATION_GUIDE_7,
    '8': MIGRATION_GUIDE_8
}


def ipa_inhibit_upgrade(ipainfo):
    """
    Create upgrade inhibitor for configured ipa-server
    """
    entries = [
        reporting.Title(
            "ipa-server does not support in-place upgrade"
        ),
        reporting.Summary(
            "An IdM server installation was detected on the system. IdM "
            "does not support in-place upgrade."
        ),
        reporting.Remediation(
            hint="Follow the IdM RHEL migration guide lines."
        ),
        reporting.ExternalLink(
            url=MIGRATION_GUIDES[get_source_major_version()],
            title="IdM migration guide",
        ),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource("package", "ipa-server"),
    ]
    return reporting.create_report(entries)


def ipa_warn_pkg_installed(ipainfo):
    """
    Warn that unused ipa-server package is installed
    """
    if ipainfo.is_client_configured:
        summary = (
            "The ipa-server package is installed but only IdM client is "
            "configured on this system."
        )
    else:
        summary = (
            "The ipa-server package is installed but neither IdM server "
            "nor client is configured on this system."
        )
    entries = [
        reporting.Title(
            "ipa-server package is installed but no IdM is configured"
        ),
        reporting.Summary(summary),
        reporting.Remediation(
            hint="Remove unused ipa-server package",
            commands=[["yum", "remove", "-y", "ipa-server"]],
        ),
        reporting.ExternalLink(
            url=MIGRATION_GUIDES[get_source_major_version()],
            title="Migrating IdM from RHEL 7 to 8",
        ),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource("package", "ipa-server"),
    ]
    return reporting.create_report(entries)



from leapp.actors import Actor
from leapp.libraries.actor import forcedefaultboot
from leapp.models import InstalledTargetKernelInfo
from leapp.tags import FinalizationPhaseTag, IPUWorkflowTag


class ForceDefaultBootToTargetKernelVersion(Actor):
    """
    Ensure the default boot entry is set to the new target kernel

    This Actor ensure that the default entry in the boot loader is set to the newly installed kernel version.
    There have been cases when the default boot entry was not set to the default kernel version. In this case the
    actor will log a warning for debugging purposes.
    """

    name = 'force_default_boot_to_target_kernel_version'
    consumes = (InstalledTargetKernelInfo,)
    produces = ()
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def process(self):
        forcedefaultboot.process()



import ast
import os

from leapp.exceptions import StopActorExecution
from leapp.libraries.common import rpms
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import CustomModifications

LEAPP_REPO_DIRS = ['/usr/share/leapp-repository']
LEAPP_PACKAGES_TO_IGNORE = ['snactor']


def _get_dirs_to_check(component):
    if component == 'repository':
        return LEAPP_REPO_DIRS
    return []


def _get_rpms_to_check(component=None):
    if component == 'repository':
        return rpms.get_leapp_packages(component=rpms.LeappComponents.REPOSITORY)
    if component == 'framework':
        return rpms.get_leapp_packages(component=rpms.LeappComponents.FRAMEWORK)
    return rpms.get_leapp_packages(components=[rpms.LeappComponents.REPOSITORY, rpms.LeappComponents.FRAMEWORK])


def deduce_actor_name(a_file):
    """
    A helper to map an actor/library to the actor name
    If a_file is an actor or an actor library, the name of the actor (name attribute of actor class) will be returned.
    Empty string is returned if the file could not be associated with any actor.
    """
    if not os.path.exists(a_file):
        return ''
    # NOTE(ivasilev) Actors reside only in actor.py files, so AST processing any other file can be skipped.
    # In case this function has been called on a non-actor file, let's go straight to recursive call on the assumed
    # location of the actor file.
    if os.path.basename(a_file) == 'actor.py':
        data = None
        with open(a_file) as f:
            try:
                data = ast.parse(f.read())
            except TypeError:
                api.current_logger().warning('An error occurred while parsing %s, can not deduce actor name', a_file)
                return ''
        # NOTE(ivasilev) Making proper syntax analysis is not the goal here, so let's get away with the bare minimum.
        # An actor file will have an Actor ClassDef with a name attribute and a process function defined
        actor = next((obj for obj in data.body if isinstance(obj, ast.ClassDef) and obj.name and
                      any(isinstance(o, ast.FunctionDef) and o.name == 'process' for o in obj.body)), None)
        # NOTE(ivasilev) obj.name attribute refers only to Class name, so for fetching name attribute need to go
        # deeper
        if actor:
            try:
                actor_name = next((expr.value.s for expr in actor.body
                                   if isinstance(expr, ast.Assign) and expr.targets[-1].id == 'name'), None)
            except (AttributeError, IndexError):
                api.current_logger().warning("Syntax Analysis for %d has failed", a_file)
                actor_name = None
            if actor_name:
                return actor_name

    # Assuming here we are dealing with a library or a file, so let's discover actor filename and deduce actor name
    # from it. Actor is expected to be found under ../../actor.py
    def _check_assumed_location(subdir):
        assumed_actor_file = os.path.join(a_file.split(subdir)[0], 'actor.py')
        if not os.path.exists(assumed_actor_file):
            # Nothing more we can do - no actor name mapping, return ''
            return ''
        return deduce_actor_name(assumed_actor_file)

    return _check_assumed_location('libraries') or _check_assumed_location('files')


def _run_command(cmd, warning_to_log, checked=True):
    """
    A helper that executes a command and returns a result or raises StopActorExecution.
    Upon success results will contain a list with line-by-line output returned by the command.
    """
    try:
        res = run(cmd, checked=checked)
        output = res['stdout'].strip()
        if not output:
            return []
        return output.split('\n')
    except CalledProcessError:
        api.current_logger().warning(warning_to_log)
        raise StopActorExecution()


def _modification_model(filename, change_type, component, rpm_checks_str=''):
    # XXX FIXME(ivasilev) Actively thinking if different model classes inheriting from CustomModifications
    # are needed or let's get away with one model for everything (as is implemented now).
    # The only difference atm is that actor_name makes sense only for repository modifications.
    return CustomModifications(filename=filename, type=change_type, component=component,
                               actor_name=deduce_actor_name(filename), rpm_checks_str=rpm_checks_str)


def check_for_modifications(component):
    """
    This will return a list of any untypical files or changes to shipped leapp files discovered on the system.
    An empty list means that no modifications have been found.
    """
    rpms = _get_rpms_to_check(component)
    dirs = _get_dirs_to_check(component)
    source_of_truth = []
    leapp_files = []
    # Let's collect data about what should have been installed from rpm
    for rpm in rpms:
        res = _run_command(['rpm', '-ql', rpm], 'Could not get a list of installed files from rpm {}'.format(rpm))
        source_of_truth.extend(res)
    # Let's collect data about what's really on the system
    for directory in dirs:
        res = _run_command(['find', directory, '-type', 'f'],
                           'Could not get a list of leapp files from {}'.format(directory))
        leapp_files.extend(res)
    # Let's check for unexpected additions
    custom_files = sorted(set(leapp_files) - set(source_of_truth))
    # Now let's check for modifications
    modified_files = []
    modified_configs = []
    for rpm in rpms:
        res = _run_command(
                ['rpm', '-V', '--nomtime', rpm], 'Could not check authenticity of the files from {}'.format(rpm),
                # NOTE(ivasilev) check is False here as in case of any changes found exit code will be 1
                checked=False)
        if res:
            api.current_logger().warning('Modifications to leapp files detected!\n%s', res)
            for modification_str in res:
                modification = tuple(modification_str.split())
                if len(modification) == 3 and modification[1] == 'c':
                    # Dealing with a configuration that will be displayed as ('S.5......', 'c', '/file/path')
                    modified_configs.append(modification)
                else:
                    # Modification of any other rpm file detected
                    modified_files.append(modification)
    return ([_modification_model(filename=f[1], component=component, rpm_checks_str=f[0], change_type='modified')
             # Let's filter out pyc files not to clutter the output as pyc will be present even in case of
             # a plain open & save-not-changed that we agreed not to react upon.
             for f in modified_files if not f[1].endswith('.pyc')] +
            [_modification_model(filename=f, component=component, change_type='custom')
             for f in custom_files] +
            [_modification_model(filename=f[2], component='configuration', rpm_checks_str=f[0], change_type='modified')
            for f in modified_configs])


def scan():
    return check_for_modifications('framework') + check_for_modifications('repository')



from leapp import reporting
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM

# Summary for mariadb-server report
report_server_inst_summary = (
    'MariaDB server component will be upgraded. Since RHEL-9 includes'
    ' MariaDB server 10.5 by default, which is incompatible with 10.3'
    ' included in RHEL-8, it is necessary to proceed with additional steps'
    ' for the complete upgrade of the MariaDB data.'
)

report_server_inst_hint = (
    'Back up your data before proceeding with the upgrade'
    ' and follow steps in the documentation section "Migrating to a RHEL 9 version of MariaDB"'
    ' after the upgrade.'
)

# Link URL for mariadb-server report
report_server_inst_link_url = 'https://access.redhat.com/articles/6743671'


def _report_server_installed():
    """
    Create report on mariadb-server package installation detection.

    Should remind user about present MariaDB server package
    installation, warn them about necessary additional steps, and
    redirect them to online documentation for the upgrade process.
    """
    reporting.create_report([
        reporting.Title('MariaDB (mariadb-server) has been detected on your system'),
        reporting.Summary(report_server_inst_summary),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.ExternalLink(title='Migrating to a RHEL 9 version of MariaDB',
                               url=report_server_inst_link_url),
        reporting.RelatedResource('package', 'mariadb-server'),
        reporting.Remediation(hint=report_server_inst_hint),
        ])


def report_installed_packages(_context=api):
    """
    Create reports according to detected MariaDB packages.

    Create the report if the mariadb-server rpm (RH signed) is installed.
    """
    has_server = has_package(DistributionSignedRPM, 'mariadb-server', context=_context)

    if has_server:
        _report_server_installed()



import pytest

from leapp.exceptions import StopActorExecution
from leapp.libraries.actor import removebootfiles
from leapp.libraries.common.testutils import logger_mocked
from leapp.libraries.stdlib import api
from leapp.models import BootContent


class remove_file_mocked(object):
    def __init__(self):
        self.called = 0
        self.files_to_remove = []

    def __call__(self, filename):
        self.called += 1
        self.files_to_remove.append(filename)


def test_remove_boot_files(monkeypatch):
    # BootContent message available
    def consume_message_mocked(*models):
        yield BootContent(kernel_path='/abc', initram_path='/def', kernel_hmac_path='/ghi')

    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_message_mocked)
    monkeypatch.setattr(removebootfiles, 'remove_file', remove_file_mocked())

    removebootfiles.remove_boot_files()

    assert removebootfiles.remove_file.files_to_remove == ['/abc', '/def', '/ghi']

    # No BootContent message available
    def consume_no_message_mocked(*models):
        yield None

    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_no_message_mocked)
    monkeypatch.setattr(removebootfiles, 'remove_file', remove_file_mocked())
    monkeypatch.setattr('leapp.libraries.stdlib.api.current_logger', logger_mocked())

    with pytest.raises(StopActorExecution):
        removebootfiles.remove_boot_files()

    assert removebootfiles.remove_file.called == 0
    assert any("Did not receive a message" in msg for msg in api.current_logger.warnmsg)


def test_remove_file_that_does_not_exist(monkeypatch):
    def remove_mocked(filepath):
        raise OSError

    monkeypatch.setattr('os.remove', remove_mocked)
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    removebootfiles.remove_file('/filepath')

    assert any("Could not remove /filepath" in msg for msg in api.current_logger.errmsg)



import errno
import os
import shutil
import tempfile


def backup_file(path):
    """ Backup the file specified by path and return the path to the backup file. """
    backup_path = '%s.leapp-backup' % path
    try:
        fd = os.open(backup_path, os.O_CREAT | os.O_EXCL, 0o600)
    except OSError as e:
        if e.errno == errno.EEXIST:
            file_name = os.path.basename(backup_path)
            fd, backup_path = tempfile.mkstemp(prefix=file_name + '.',
                                               dir=os.path.dirname(backup_path))
        else:
            raise
    os.close(fd)

    shutil.copyfile(path, backup_path)
    return backup_path



import errno
import os
import shutil

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import InitrdIncludes  # deprecated
from leapp.models import InstalledTargetKernelInfo, TargetInitramfsTasks
from leapp.utils.deprecation import suppress_deprecation

DRACUT_DIR = '/usr/lib/dracut/modules.d/'


def _get_target_kernel_modules_dir(kernel_version):
    """
    Return the path where the custom kernel modules should be copied.
    """

    modules_dir = os.path.join('/', 'lib', 'modules', kernel_version, 'extra', 'leapp')

    return modules_dir


def _copy_modules(modules, dst_dir, kind):
    """
    Copy modules of given kind to the specified destination directory.

    Attempts to remove an cleanup by removing the existing destination
    directory. If the directory does not exist, it is created anew. Then, for
    each module message, it checks if the module has a module path specified. If
    the module already exists in the destination directory, a debug message is
    logged, and the operation is skipped. Otherwise, the module is copied to the
    destination directory.

    """

    try:
        os.makedirs(dst_dir)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(dst_dir):
            pass
        else:
            raise

    for module in modules:
        if not module.module_path:
            continue

        dst_path = os.path.join(dst_dir, os.path.basename(module.module_path))
        if os.path.exists(dst_path):
            api.current_logger().debug(
                'The {name} {kind} module has been already installed. Skipping.'
                .format(name=module.name, kind=kind))
            continue

        copy_fn = shutil.copytree
        if os.path.isfile(module.module_path):
            copy_fn = shutil.copy2

        try:
            api.current_logger().debug(
                'Copying {kind} module "{name}" to "{path}".'
                .format(kind=kind, name=module.name, path=dst_path))

            copy_fn(module.module_path, dst_path)
        except shutil.Error as e:
            api.current_logger().error(
                    'Failed to copy {kind} module "{name}" from "{source}" to "{target}"'.format(
                        kind=kind, name=module.name, source=module.module_path, target=dst_dir),
                    exc_info=True)
            raise StopActorExecutionError(
                message='Failed to install {kind} modules required in the initram. Error: {error}'.format(
                    kind=kind, error=str(e))
            )


@suppress_deprecation(InitrdIncludes)
def _get_files():
    files = {f for i in api.consume(InitrdIncludes) for f in i.files}
    files.update([f for i in api.consume(TargetInitramfsTasks) for f in i.include_files])
    return files


def _get_modules():
    # NOTE(pstodulk): Duplicated tasks are not filtered out, nor checked in the actor.
    # Currently possible conflicting tasks are detected by the checkinitramfstasks
    # actor that inhibits the upgrade if any conflicts are detected. User is
    # supposed to create any such tasks before the reporting phase, so we
    # are able to check it.
    #
    modules = {'dracut': [], 'kernel': []}
    for task in api.consume(TargetInitramfsTasks):
        modules['dracut'].extend(task.include_dracut_modules)
        modules['kernel'].extend(task.include_kernel_modules)

    return modules


def process():
    files = _get_files()
    modules = _get_modules()

    if not files and not modules['kernel'] and not modules['dracut']:
        api.current_logger().debug(
            'No additional files or modules required to add into the target initramfs.')
        return

    target_kernel_info = next(api.consume(InstalledTargetKernelInfo), None)
    if not target_kernel_info:
        raise StopActorExecutionError(
            'Cannot get version of the installed RHEL-8 kernel',
            details={'Problem': 'Did not receive a message with installed RHEL-8 kernel version'
                                ' (InstalledTargetKernelVersion)'})

    _copy_modules(modules['dracut'], DRACUT_DIR, 'dracut')
    _copy_modules(modules['kernel'], _get_target_kernel_modules_dir(target_kernel_info.uname_r), 'kernel')

    # Discover any new modules and regenerate modules.dep
    should_regenerate = any(module.module_path is not None for module in modules['kernel'])
    if should_regenerate:
        try:
            run(['depmod', target_kernel_info.uname_r, '-a'])
        except CalledProcessError as e:
            raise StopActorExecutionError('Failed to generate modules.dep and map files.', details={'details': str(e)})

    try:
        # multiple files|modules need to be quoted, see --install | --add in dracut(8)
        dracut_module_names = list({module.name for module in modules['dracut']})
        kernel_module_names = list({module.name for module in modules['kernel']})
        cmd = ['dracut', '-f', '--kver', target_kernel_info.uname_r]
        if files:
            cmd += ['--install', '{}'.format(' '.join(files))]
        if modules['dracut']:
            cmd += ['--add', '{}'.format(' '.join(dracut_module_names))]
        if modules['kernel']:
            cmd += ['--add-drivers', '{}'.format(' '.join(kernel_module_names))]

        run(cmd)
    except CalledProcessError as e:
        # just hypothetic check, it should not die
        raise StopActorExecutionError('Cannot regenerate dracut image.', details={'details': str(e)})



from leapp.actors import Actor
from leapp.models import DNFWorkaround
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class RegisterYumAdjustment(Actor):
    """
    Registers a workaround which will adjust the yum directories during the upgrade.
    """

    name = 'register_yum_adjustment'
    consumes = ()
    produces = (DNFWorkaround,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        self.produce(
            DNFWorkaround(
                display_name='yum config fix',
                script_path=self.get_tool_path('handleyumconfig'),
            )
        )



from leapp.models import CustomTargetRepository, fields, Model
from leapp.topics import SystemFactsTopic


class TargetOSInstallationImage(Model):
    """
    An installation image of a target OS requested to be the source of target OS packages.
    """
    topic = SystemFactsTopic
    path = fields.String()
    mountpoint = fields.String()
    repositories = fields.List(fields.Model(CustomTargetRepository))
    rhel_version = fields.String(default='')
    was_mounted_successfully = fields.Boolean(default=False)



import logging

import pytest
import six

from leapp.libraries.actor import sctpdlm

if six.PY2:
    from mock import mock_open
else:
    from unittest.mock import mock_open


# TODO Confirm with the team the way to mock builtin open
#   and apply this throughout the repo


@pytest.mark.parametrize(
    ('config', 'open_raises', 'exp_return',),
    [
        ('', IOError, False),
        ('', OSError, False),
        ('log_debug=1\npost_join_delay=10', None, False),
        ('log_debug=1\npost_join_delay=10\nprotocol=sctp', None, True),
        ('log_debug=1\npost_join_delay=10\nprotocol=detect', None, True),
        ('log_debug=1\npost_join_delay=10\nprotocol=1', None, True),
        ('log_debug=1\npost_join_delay=10\nprotocol=2', None, True),
        ('log_debug=1\npost_join_delay=10\nprotocol=tcp', None, False),
        ('log_debug=1\npost_join_delay=10', None, False),
    ],
)
def test_check_dlm_cfgfile(config, open_raises, exp_return):
    if open_raises:
        mock_open.side_effect = open_raises
    assert (
        sctpdlm.check_dlm_cfgfile(_open=mock_open(read_data=config))
        == exp_return
    )


@pytest.mark.parametrize(
    ('config', 'open_raises', 'exp_return'),
    [
        ('', IOError, False),
        ('', OSError, False),
        ('DLM_CONTROLD_OPTS="- f 0 -q 0 --protocol=sctp"', None, True),
        ('DLM_CONTROLD_OPTS="- f 0 -q 0 -r detect"', None, True),
        ('DLM_CONTROLD_OPTS="- f 0 -q 0 --protocol tcp"', None, False),
    ],
)
def test_check_dlm_sysconfig(config, open_raises, exp_return):
    if open_raises:
        mock_open.side_effect = open_raises
    assert (
        sctpdlm.check_dlm_sysconfig(_open=mock_open(read_data=config))
        == exp_return
    )


@pytest.mark.parametrize(
    (
        'check_dlm_cfg_file_returns',
        'check_dlm_sysconfig_returns',
        'exp_return',
        'text_in_log',
    ),
    [
        (True, False, True, 'DLM is configured to use SCTP on dlm.conf.'),
        (False, True, True, 'DLM is configured to use SCTP on sysconfig.'),
        (False, False, False, ''),
    ],
)
def test_is_dlm_using_sctp(
    check_dlm_cfg_file_returns,
    check_dlm_sysconfig_returns,
    exp_return,
    text_in_log,
    monkeypatch,
    caplog,
):
    monkeypatch.setattr(
        sctpdlm, 'check_dlm_cfgfile', lambda: check_dlm_cfg_file_returns
    )
    monkeypatch.setattr(
        sctpdlm, 'check_dlm_sysconfig', lambda: check_dlm_sysconfig_returns
    )
    with caplog.at_level(logging.DEBUG):
        assert sctpdlm.is_dlm_using_sctp() == exp_return
    if text_in_log:
        assert text_in_log in caplog.text
